#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import queue
import json
import base64
import socket
import time
import random
import urllib.request
import urllib.parse
import urllib.error
from enum import Enum
from enum import IntEnum
from hub.utils.codec import Codec
from hub.utils.providers import TopicProvider
from hub.utils.providers import DeviceInfoProvider
from hub.utils.providers import ConnClientProvider
from hub.utils.providers import LoggerProvider
from hub.manager.manager import TaskManager
from hub.services.gateway.gateway import Gateway
from hub.services.rrpc.rrpc import Rrpc
from hub.services.broadcast.broadcast import Broadcast
from hub.services.shadow.shadow import Shadow
from hub.services.ota.ota import Ota

class SingletonType(type):
    _instance_lock = threading.Lock()
    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            with SingletonType._instance_lock:
                if not hasattr(cls, "_instance"):
                    cls._instance = super(SingletonType,cls).__call__(*args, **kwargs)
        return cls._instance

class QcloudHub(metaclass=SingletonType):
    """
    使用单例模式构建,保证对象只有一份
    """
    def __init__(self, device_file, userdata=None, tls=True, domain=None, useWebsocket=False):
        self.hub = QcloudHubProvider(device_file, userdata=userdata, tls=tls, domain=domain, useWebsocket=useWebsocket)

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

class QcloudHubProvider(object):
    """事件核心处理层
    作为explorer/user层与协议层的中间层,负责上下层通道建立、消息分发等事物
    """
    def __init__(self, device_file, userdata=None, tls=True, domain=None, useWebsocket=False):
        self.__tls = tls
        self.__useWebsocket = useWebsocket
        self.__key_mode = True
        self.__userdata = userdata
        self.__provider = None
        self.__protocol = None
        self.__domain = domain
        self.__host = ""
        self.__log_provider = LoggerProvider()
        self._logger = self.__log_provider.logger
        self.__codec = Codec()
        self.__gateway = None
        self.__device_info = DeviceInfoProvider(device_file)

        self.__hub_state = self.HubState.INITIALIZED
        self._topic = TopicProvider(self.__device_info.product_id, self.__device_info.device_name)

        self.__ntp_lock = threading.Lock()
        self.__ntptime = self.NtpTime()

        """存放explorer层注册到hub层的回调函数
        只存放explorer层独有的功能所需的回调(诸如数据模板),
        类似on_connect等explorer和用户层都可能注册的回调在hub层使用专门的函数与之对应
        """
        self.__explorer_callback = {}

        """ 存放用户注册的回调函数 """
        self.__user_callback = {}

        """
        保存__on_subscribe()返回的mid和qos对,用以判断订阅是否成功
        """
        self.__subscribe_res = {}

        self.__ota_map = {}
        self.__rrpc_map = {}
        self.__shadow_map = {}
        self.__broadcast_map = {}

        self.__user_topics = []
        self.__user_topics_subscribe_request = {}
        self.__user_topics_unsubscribe_request = {}
        self.__user_topics_request_lock = threading.Lock()
        self.__user_topics_unsubscribe_request_lock = threading.Lock()

        self.__loop_worker = self.LoopWorker()
        self.__event_worker = self.EventWorker()
        self.__register_event_callback()

        """
        hub层注册到mqtt的回调
        """
        self.__user_on_connect = None
        self.__user_on_disconnect = None
        self.__user_on_publish = None
        self.__user_on_subscribe = None
        self.__user_on_unsubscribe = None
        self.__user_on_message = None

        self.__protocol_init(domain, useWebsocket)

    class HubState(Enum):
        """ 连接状态 """
        INITIALIZED = 1
        CONNECTING = 2
        CONNECTED = 3
        DISCONNECTING = 4
        DISCONNECTED = 5
        DESTRUCTING = 6
        DESTRUCTED = 7
    
    class ErrorCode(Enum):
        ERR_NONE = 0  # 成功
        ERR_TOPIC_NONE = -1000  # topic为空
        
    
    class StateError(Exception):
        def __init__(self, err):
            Exception.__init__(self, err)

    class NtpTime(object):
        def __init__(self):
            self._ntp_recvied = False
            """
            ntp请求的发送和接收时间戳
            """
            self._ntp_send_timestamp = 0
            self._ntp_recv_timestamp = 0
            """
            从平台获取的ntptime时间戳
            """
            self._ntptime1 = 0
            self._ntptime2 = 0

    class LoggerLevel(Enum):
        INFO = "info"
        DEBUG = "debug"
        WARNING = "warring"
        ERROR = "error"

    class device_property(object):
        def __init__(self):
            self.key = ""
            self.data = ""
            self.data_buff_len = 0
            self.type = ""

    # 管理连接相关资源
    class LoopWorker(object):
        """ mqtt连接管理维护 """
        def __init__(self):
            self._connect_async_req = False
            self._exit_req = True
            self._runing_state = False
            self._exit_req_lock = threading.Lock()
            self._thread = TaskManager.LoopThread()

    class EventWorker(object):
        """ 事件管理 """
        def __init__(self):
            self._thread = TaskManager.EventCbThread()
        
        def _register_event_callback(self, connect, disconnect,
                                        message, publish, subscribe, unsubscribe):
            self._thread.register_event_callback(self.EventPool.CONNECT, connect)
            self._thread.register_event_callback(self.EventPool.DISCONNECT, disconnect)
            self._thread.register_event_callback(self.EventPool.MESSAGE, message)
            self._thread.register_event_callback(self.EventPool.PUBLISH, publish)
            self._thread.register_event_callback(self.EventPool.SUBSCRISE, subscribe)
            self._thread.register_event_callback(self.EventPool.UNSUBSCRISE, unsubscribe)

            self._thread.start()

        class EventPool(object):
            CONNECT = "connect"
            DISCONNECT = "disconnect"
            MESSAGE = "message"
            PUBLISH = "publish"
            SUBSCRISE = "subscribe"
            UNSUBSCRISE = "unsubscribe"

    class sReplyPara(object):
        def __init__(self):
            self.timeout_ms = 0
            self.code = -1
            self.status_msg = None

    def __assert(self, param):
        if param is None or len(param) == 0:
            raise ValueError('Invalid param.')

    def __register_event_callback(self):
        self.__event_worker._register_event_callback(self.__user_connect,
                                                    self.__user_disconnect,
                                                    self.__user_message,
                                                    self.__user_publish,
                                                    self.__user_subscribe,
                                                    self.__user_unsubscribe)

    """
    处理用户回调
    基于explorer接入时会将用户回调赋值到本层用户回调函数
    """
    def __user_connect(self, value):
        # client, user_data, session_flag, rc = value
        session_flag, rc = value
        if self.__user_on_connect is not None:
            try:
                self.__user_on_connect(session_flag['session present'], rc, self.__userdata)
            except Exception as e:
                self._logger.error("on_connect process raise exception:%r" % e)
        pass

    def __user_disconnect(self, value):
        self.__user_on_disconnect(value, self.__userdata)
        pass

    def __user_publish(self, value):
        self.__user_on_publish(value, self.__userdata)
        pass

    def __user_subscribe(self, value):
        qos, mid = value
        self.__user_on_subscribe(qos, mid, self.__userdata)
        pass

    def __user_unsubscribe(self, value):
        self.__user_on_unsubscribe(value, self.__userdata)
        pass

    def __user_message(self, value):
        message = value
        topic = message.topic
        qos = message.qos
        payload = json.loads(message.payload.decode('utf-8'))
        # print(">>>>>>> from qcloud:%s, topic:%s" % (payload, topic))

        topic_prefix = topic[0:topic.find("/")]

        pos = topic.rfind("/")
        device_name = topic[pos + 1:len(topic)]

        topic_split = topic[0:pos]
        pos = topic_split.rfind("/")
        product_id = topic_split[pos + 1:len(topic_split)]
        client = product_id + device_name

        if topic_prefix == "$thing" or topic_prefix == "$template":
            if topic == self._topic.template_raw_topic_sub:
                # 调用explorer向hub注册的回调处理
                self._logger.info("Reserved: template raw topic")

            elif topic == self._topic.template_topic_sub:
                if self.__user_callback[topic] is not None:
                    self.__user_callback[topic](topic, qos, payload, self.__userdata)
                else:
                    self._logger.error("no callback for topic %s" % topic)
            else:
                if self.__explorer_callback[topic] is not None:
                    self.__explorer_callback[topic](topic, qos, payload, self.__userdata)
                else:
                    self._logger.error("no callback for topic %s" % topic)

        elif topic_prefix == "$sys":
            # 获取时间作为内部服务,不通知到用户
            if (('type' in payload and payload["type"] == "get") and
                ('time' in payload and payload["time"] is not None)):
                    self.__ntptime._ntp_recvied = True
                    self.__ntptime._ntp_recv_timestamp = int(time.time() * 1000)
                    self.__ntptime._ntptime1 = payload["ntptime1"]
                    self.__ntptime._ntptime2 = payload["ntptime2"]
            else:
                self.__user_on_message(topic, qos, payload, self.__userdata)

        elif topic_prefix == "$gateway":
            self.__gateway.handle_gateway(topic, payload)

        elif topic_prefix == "$ota":
            if (client not in self.__ota_map.keys()
                    or self.__ota_map[client] is None):
                self._logger.error("[template] not found template handle for client:%s" % (client))
                return None

            ota = self.__ota_map[client]
            ota.handle_ota(topic, qos, payload, self.__userdata)

        elif topic_prefix == "$rrpc":
            # topic:$rrpc/rxd/${productID}/${deviceName}/${processID}
            pos = topic.rfind("/")
            topic_split1 = topic[0:pos]

            pos = topic_split1.rfind("/")
            device_name = topic_split1[pos + 1:len(topic_split1)]

            topic_split = topic_split1[0:pos]
            pos = topic_split.rfind("/")
            product_id = topic_split[pos + 1:len(topic_split)]

            client = product_id + device_name
            if (client not in self.__rrpc_map.keys()
                    or self.__rrpc_map[client] is None):
                self._logger.error("[template] not found template handle for client:%s" % (client))
                return None

            rrpc = self.__rrpc_map[client]
            rrpc.handle_rrpc(topic, qos, payload, self.__userdata)
        
        elif topic_prefix == "$shadow":
            if (client not in self.__shadow_map.keys()
                    or self.__shadow_map[client] is None):
                self._logger.error("[template] not found template handle for client:%s" % (client))
                return None

            shadow = self.__shadow_map[client]
            shadow.handle_shadow(topic, qos, payload, self.__userdata)
        
        elif topic_prefix == "$broadcast":
            if (client not in self.__broadcast_map.keys()
                    or self.__broadcast_map[client] is None):
                self._logger.error("[template] not found template handle for client:%s" % (client))
                return None

            broadcast = self.__broadcast_map[client]
            broadcast.handle_broadcast(topic, qos, payload, self.__userdata)

        elif topic in self.__user_topics:
            if self.__user_callback[topic] is not None:
                self.__user_callback[topic](topic, qos, payload, self.__userdata)
            else:
                self._logger.error("no callback for topic %s" % topic)

        else:
            if self.__explorer_callback[topic] is not None:
                self.__explorer_callback[topic](topic, qos, payload, self.__userdata)
            else:
                self._logger.error("unknow topic:%s" % topic)
        pass

    def __on_connect(self, client, user_data, session_flag, rc):
        if rc == 0:
            self.__protocol.reset_reconnect_wait()
            self.__hub_state = self.HubState.CONNECTED
            self.__event_worker._thread.post_message(self.__event_worker.EventPool.CONNECT, (session_flag, rc))

        pass

    def __on_disconnect(self, client, user_data, rc):
        if self.__hub_state == self.HubState.DISCONNECTING:
            self.__hub_state = self.HubState.DISCONNECTED
        elif self.__hub_state == self.HubState.DESTRUCTING:
            self.__hub_state = self.HubState.DESTRUCTED
        elif self.__hub_state == self.HubState.CONNECTED:
            self.__hub_state = self.HubState.DISCONNECTED
        else:
            self._logger.error("state error:%r" % self.__hub_state)
            return

        self.__user_topics_subscribe_request.clear()
        self.__user_topics_unsubscribe_request.clear()
        self.__user_topics.clear()

        if self.__gateway is not None:
            self.__gateway.gateway_reset()

        """
        将disconnect事件通知到explorer
        """
        ex_topic = "$explorer/from/disconnect"
        if ex_topic in self.__explorer_callback:
            if self.__explorer_callback[ex_topic] is not None:
                self.__explorer_callback[ex_topic](client, self.__userdata, rc)
            else:
                self._logger.error("no callback for topic %s" % ex_topic)

        self.__event_worker._thread.post_message(self.__event_worker.EventPool.DISCONNECT, (rc))
        if self.__hub_state == self.HubState.DESTRUCTED:
            self.__event_worker._thread.stop()

    def __on_message(self, client, user_data, message):
        self.__event_worker._thread.post_message(self.__event_worker.EventPool.MESSAGE, (message))

    def __on_publish(self, client, user_data, mid):
        self.__event_worker._thread.post_message(self.__event_worker.EventPool.PUBLISH, (mid))

    def __on_subscribe(self, client, user_data, mid, granted_qos):
        qos = granted_qos[0]
        # todo:mid,qos
        self.__subscribe_res[mid] = qos
        self.__event_worker._thread.post_message(self.__event_worker.EventPool.SUBSCRISE, (qos, mid))

    def __on_unsubscribe(self, client, user_data, mid):
        self.__event_worker._thread.post_message(self.__event_worker.EventPool.UNSUBSCRISE, (mid))

    def _loop(self):
        if self.__hub_state not in (self.HubState.INITIALIZED,
                                     self.HubState.DISCONNECTED):
            raise self.StateError("current state is not in INITIALIZED or DISCONNECTED")
        self.__hub_state = self.HubState.CONNECTING

        if self.__protocol.connect() is not True:
            self.__hub_state = self.HubState.INITIALIZED
            return

        while True:
            if self.__loop_worker._exit_req:
                if self.__hub_state == self.HubState.DESTRUCTING:
                    self.__loop_worker._thread.stop()
                    self.__hub_state = self.HubState.DESTRUCTED
                break
            try:
                self.__hub_state = self.HubState.CONNECTING
                """
                实际连接
                """
                self.__protocol.reconnect()
            except (socket.error, OSError) as e:
                self._logger.error("mqtt reconnect error:" + str(e))
                # 失败处理 待添加
                if self.__hub_state == self.HubState.CONNECTING:
                    self.__hub_state = self.HubState.DISCONNECTED
                    self.__protocol.reset_reconnect_wait()

                    if self.__hub_state == self.HubState.DESTRUCTING:
                        self.__loop_worker._thread.stop()
                        self.__hub_state = self.HubState.DESTRUCTED
                        break
                    self.__protocol.reconnect_wait()
                continue
            """
            调用循环调用mqtt loop读取消息
            """
            self.__protocol.loop()
            """
            mqtt loop接口失败(异常导致的disconnect)
            1.将disconnect事件通知到用户
            2.清理sdk相关资源
            """
            if self.__hub_state == self.HubState.CONNECTED:
                self.__on_disconnect(None, None, -1)
            """
            清理线程资源
            """
            if self.__loop_worker._exit_req:
                if self.__hub_state == self.HubState.DESTRUCTING:
                    self.__loop_worker._thread.stop()
                    self.__hub_state = self.HubState.DESTRUCTED
                break
            self.__protocol.reconnect_wait()
        pass

    def registerMqttCallback(self, on_connect, on_disconnect,
                            on_message, on_publish,
                            on_subscribe, on_unsubscribe):
        """Register user mqtt callback

        Register user mqtt callback for mqtt
        Args:
            on_connect: mqtt connect callback
            on_disconnect: mqtt disconnect callback
            on_message: mqtt message callback
            on_publish: mqtt publish callback
            on_subscribe: mqtt subscribe callback
            on_unsubscribe: mqtt unsubscribe callback
        Returns:
            success: default
            fail: default
        """
        self.__user_on_connect = on_connect
        self.__user_on_disconnect = on_disconnect
        self.__user_on_message = on_message
        self.__user_on_publish = on_publish
        self.__user_on_subscribe = on_subscribe
        self.__user_on_unsubscribe = on_unsubscribe

    def registerUserCallback(self, topic, callback):
        """Register user callback

        Register user callback for a topic
        Args:
            topic: topic
            callback: user callback
        Returns:
            success: default
            fail: default
        """
        self.__user_topics.append(topic)
        self.__user_callback[topic] = callback

    def isMqttConnected(self):
        """Is mqtt connected

        Is mqtt connected
        Args: None
        Returns:
            success: True/False
        """
        return self.__hub_state == self.HubState.CONNECTED

    def getConnectState(self):
        """Get connect state

        Get device current connect state
        Args: None
        Returns:
            success: connect state
        """
        return self.__hub_state

    def register_explorer_callback(self, topic, callback):
        """
        注册explorer层回调到hub层
        topic可为单个topic或topic列表
        """
        if isinstance(topic, str):
            if topic is not None or len(topic) > 0:
                self.__explorer_callback[topic] = callback
        if isinstance(topic, list):
            for tp in topic:
                self.__explorer_callback[tp] = callback
        pass

    def __protocol_reinit(self):
        self.__protocol_init(self.__domain, self.__useWebsocket)

    # 连接协议(mqtt/websocket)初始化
    def __protocol_init(self, domain=None, useWebsocket=False):
        auth_mode = self.__device_info.auth_mode
        device_name = self.__device_info.device_name
        product_id = self.__device_info.product_id
        device_secret = self.__device_info.device_secret
        ca = self.__device_info.ca_file
        cert = self.__device_info.cert_file
        key = self.__device_info.private_key_file

        """
        由于ConnClientProvider是单例模式,因此没有device secret不能创建protocol对象
        """
        if device_secret == "YOUR_DEVICE_SECRET":
            self._logger.error("device secret invalid!")
            return

        if useWebsocket is False:
            if domain is None or domain == "":
                self.__host = product_id + ".iotcloud.tencentdevices.com"
            else:
                self.__host = product_id + domain
        else:
            self.__host = product_id + ".ap-guangzhou.iothub.tencentdevices.com"

        self.__provider = ConnClientProvider(self.__host, product_id, device_name, device_secret,
                                                websocket=useWebsocket, tls=self.__tls, logger=self._logger)
        self.__protocol = self.__provider.protocol

        if auth_mode == "CERT":
            self.__protocol.set_cert_file(ca, cert, key)
        
        self.__protocol.register_event_callbacks(self.__on_connect,
                                                    self.__on_disconnect,
                                                    self.__on_message,
                                                    self.__on_publish,
                                                    self.__on_subscribe,
                                                    self.__on_unsubscribe)
        pass

    def setReconnectInterval(self, max_sec, min_sec):
        """Set mqtt reconnect interval

        Set mqtt reconnect interval
        Args:
            max_sec: reconnect max time
            min_sec: reconnect min time
        Returns:
            success: default
            fail: default
        """
        if self.__protocol is None:
            self._logger.error("Set failed: client is None")
            return
        self.__protocol.set_reconnect_interval(max_sec, min_sec)
        self.__protocol.config_connect()

    def setMessageTimout(self, timeout):
        """Set message overtime time

        Set message overtime time
        Args:
            timeout: mqtt keepalive value
        Returns:
            success: default
            fail: default
        """
        if self.__protocol is None:
            self._logger.error("Set failed: client is None")
            return
        self.__protocol.set_message_timout(timeout)

    def setKeepaliveInterval(self, interval):
        """Set mqtt keepalive interval

        Set mqtt keepalive interval
        Args:
            interval: mqtt keepalive interval
        Returns:
            success: default
            fail: default
        """
        if self.__protocol is None:
            self._logger.error("Set failed: client is None")
            return
        self.__protocol.set_keepalive_interval(interval)

    def getProductID(self):
        """Get product id

        Get product id
        Args: None
        Returns:
            success: product id
            fail: None
        """
        return self.__device_info.product_id

    def getDeviceName(self):
        """Get device name

        Get device name
        Args: None
        Returns:
            success: device name
            fail: None
        """
        return self.__device_info.device_name

    def getNtpAccurateTime(self):
        """Get NTP time

        Get NTP time
        Args: None
        Returns:
            success: device current accurate timestamp
            fail: -1
        """
        timestamp = -1

        with self.__ntp_lock:
            sys_topic_sub = self._topic.sys_topic_sub
            sys_topic_pub = self._topic.sys_topic_pub
            rc, mid = self.subscribe(sys_topic_sub, 0)
            if rc == 0:
                payload = {
                    "type": "get",
                    "resource": [
                        "time"
                    ],
                }

                self.__ntptime._ntp_send_timestamp = int(time.time() * 1000)
                self.publish(sys_topic_pub, payload, 0)
            else:
                self._logger.error("[sys] subscribe error:rc:%d,topic:%s" % (rc, sys_topic_sub))
                self.unsubscribe(sys_topic_sub)
                return timestamp

            cnt = 0
            while cnt < 3:
                if self.__ntptime._ntp_recvied is True:
                    timestamp = (self.__ntptime._ntptime1 + self.__ntptime._ntptime2 +
                                    self.__ntptime._ntp_recv_timestamp - self.__ntptime._ntp_send_timestamp) / 2
                    break
                pass
                time.sleep(0.5)
                cnt += 1

        self.unsubscribe(sys_topic_sub)
        return timestamp

    def connect(self):
        """Connect

        Device connect
        Args: None
        Returns:
            success: thread start result
            fail: thread start result
        """
        self.__loop_worker._connect_async_req = True
        with self.__loop_worker._exit_req_lock:
            self.__loop_worker._exit_req = False
        return self.__loop_worker._thread.start(self._loop)

    def disconnect(self):
        """Disconnect

        Device disconnect
        Args: None
        Returns:
            success: default
            fail: default
        """
        self._logger.debug("disconnect")
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")
        self.__hub_state = self.HubState.DISCONNECTING
        if self.__loop_worker._connect_async_req is True:
            with self.__loop_worker._exit_req_lock:
                self.__loop_worker._exit_req = True

        self.__protocol.disconnect()
        self.__loop_worker._thread.stop()

    def subscribe(self, topic, qos):
        """Subscribe topic

        Subscribe topic
        Args:
            topic: topic
            qos: mqtt qos
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")
        if isinstance(topic, tuple):
            topic, qos = topic
        if isinstance(topic, str):
            self.__user_topics_request_lock.acquire()
            rc, mid = self.__protocol.subscribe(topic, qos)
            if rc == 0:
                self.__user_topics_subscribe_request[mid] = [(topic, qos)]
            self.__user_topics_request_lock.release()
            return rc, mid
        # topic format [(topic1, qos),(topic2,qos)]
        if isinstance(topic, list):
            self.__user_topics_request_lock.acquire()
            rc, mid = self.__protocol.subscribe(topic)
            if rc == 0:
                self.__user_topics_subscribe_request[mid] = [topic]
            self.__user_topics_request_lock.release()
            return rc, mid
        pass

    def unsubscribe(self, topic):
        """Unsubscribe topic

        Unsubscribe topic what is subscribed
        Args:
            topic: topic
        Returns:
            success: zero and unsubscribe mid
            fail: negative number and unsubscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")
        unsubscribe_topics = []
        if topic is None or len(topic) == 0:
            raise ValueError('Invalid topic.')
        if isinstance(topic, str):
            # topic判断
            unsubscribe_topics.append(topic)
        elif isinstance(topic, list):
            for tp in topic:
                unsubscribe_topics.append(tp)
            pass
        with self.__user_topics_unsubscribe_request_lock:
            if len(unsubscribe_topics) == 0:
                return self.ErrorCode.ERR_TOPIC_NONE, -1
            rc, mid = self.__protocol.unsubscribe(unsubscribe_topics)
            if rc == 0:
                self.__user_topics_unsubscribe_request[mid] = unsubscribe_topics
            return rc, mid
        pass

    def publish(self, topic, payload, qos):
        """Publish message

        Publish message
        Args:
            topic: topic
            payload: publish message
            qos: mqtt qos
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")
        if topic is None or len(topic) == 0:
            raise ValueError('Invalid topic.')
        if qos < 0:
            raise ValueError('Invalid QoS level.')

        return self.__protocol.publish(topic, json.dumps(payload), qos)
    
    def dynregDevice(self, timeout=10):
        """Dynamic register

        Get the device secret from the Cloud
        Args:
            timeout: request timeout
        Returns:
            success: return zero and device secret
            fail: -1 and error message
        """
        sign_format = '%s\n%s\n%s\n%s\n%s\n%d\n%d\n%s'
        url_format = '%s://ap-guangzhou.gateway.tencentdevices.com/device/register'
        request_format = "{\"ProductId\":\"%s\",\"DeviceName\":\"%s\"}"

        device_name = self.__device_info.device_name
        product_id = self.__device_info.product_id
        product_secret = self.__device_info.product_secret

        request_text = request_format % (product_id, device_name)
        request_hash = self.__codec.Hash.sha256_encode(request_text.encode("utf-8"))

        nonce = random.randrange(2147483647)
        timestamp = int(time.time())
        sign_content = sign_format % (
            "POST", "ap-guangzhou.gateway.tencentdevices.com",
            "/device/register", "", "hmacsha256", timestamp,
            nonce, request_hash)
        sign_base64 = self.__codec.Base64.encode(self.__codec.Hmac.sha256_encode(product_secret.encode("utf-8"),
                            sign_content.encode("utf-8")))

        header = {
            'Content-Type': 'application/json; charset=utf-8',
            "X-TC-Algorithm": "hmacsha256",
            "X-TC-Timestamp": timestamp,
            "X-TC-Nonce": nonce,
            "X-TC-Signature": sign_base64
        }
        data = bytes(request_text, encoding='utf-8')

        context = None
        if self.__tls:
            request_url = url_format % 'https'
            context = self.__codec.Ssl().create_content()
        else:
            request_url = url_format % 'http'
        self._logger.info('dynreg url {}'.format(request_url))
        req = urllib.request.Request(request_url, data=data, headers=header)
        with urllib.request.urlopen(req, timeout=timeout, context=context) as url_file:
            reply_data = url_file.read().decode('utf-8')
            reply_obj = json.loads(reply_data)
            resp = reply_obj['Response']
            if 'Len' in resp and resp['Len'] > 0:
                reply_obj_data = reply_obj['Response']["Payload"]
                if reply_obj_data is not None:
                    psk = self.__codec._AESUtil.decrypt(reply_obj_data.encode('UTF-8') , product_secret[:self.__codec._AESUtil.BLOCK_SIZE_16].encode('UTF-8'),
                                        '0000000000000000'.encode('UTF-8'))
                    psk = psk.decode('UTF-8', 'ignore').strip().strip(b'\x00'.decode())
                    user_dict = json.loads(psk)
                    self._logger.info('encrypt type: {}'.format(
                        user_dict['encryptionType']))

                    self.__device_info.update_config_file(user_dict['psk'])
                    self.__protocol_reinit()

                    return 0, user_dict['psk']
                else:
                    self._logger.warring('payload is null')
                    return -1, 'payload is null'
            else:
                err_code = resp['Error']
                self._logger.error('code: {}, error message: {}'.format(
                    err_code, err_code['Message']))
                return -1, err_code['Message']

    def isSubdevStatusOnline(self, sub_productId, sub_devName):
        """Sub-device status

        Determine if the device is online
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: if device is online
        """
        return self.__gateway.is_subdev_status_online(sub_productId, sub_devName)

    def updateSubdevStatus(self, sub_productId, sub_devName, status):
        """Update device status

        Update sub-device local status
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            status: new status
        Returns:
            success: None
        """
        return self.__gateway.update_subdev_status(sub_productId, sub_devName, status)

    def gatewaySubdevGetConfigList(self):
        """Get sub-device list

        Get the list of sub-devices in the configuration file
        Args: None
        Returns:
            success: sub-device list
        """
        return self.__gateway.gateway_get_subdev_config_list()

    def gatewaySubdevOnline(self, sub_productId, sub_devName):
        """Make sub-device online

        Make sub-device online
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: zero and publish mid
        """
        self.__assert(sub_productId)
        self.__assert(sub_devName)

        gateway_topic_pub = self._topic.gateway_topic_pub
        return self.__gateway.gateway_subdev_online(gateway_topic_pub, 0, sub_productId, sub_devName)

    def gatewaySubdevOffline(self, sub_productId, sub_devName):
        """Make sub-device offline

        Make sub-device offline
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: zero and publish mid
        """
        self.__assert(sub_productId)
        self.__assert(sub_devName)

        gateway_topic_pub = self._topic.gateway_topic_pub
        return self.__gateway.gateway_subdev_offline(gateway_topic_pub, 0, sub_productId, sub_devName)

    def gatewaySubdevBind(self, sub_productId, sub_devName, sub_secret):
        """Bind sub-device

        Gateway device bind sub-device
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            sub_secret: sub-device secret
        Returns:
            success: zero and publish mid
        """
        self.__assert(sub_productId)
        self.__assert(sub_devName)
        self.__assert(sub_secret)

        gateway_topic_pub = self._topic.gateway_topic_pub
        return self.__gateway.gateway_subdev_bind(gateway_topic_pub, 0, sub_productId, sub_devName, sub_secret)

    def gatewaySubdevUnbind(self, sub_productId, sub_devName):
        """Unbind sub-device

        Gateway device unbind sub-device
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            sub_secret: sub-device secret
        Returns:
            success: zero and publish mid
        """
        self.__assert(sub_productId)
        self.__assert(sub_devName)

        gateway_topic_pub = self._topic.gateway_topic_pub
        return self.__gateway.gateway_subdev_unbind(gateway_topic_pub, 0, sub_productId, sub_devName)

    def gatewaySubdevGetBindList(self, product_id, device_name):
        """Get sub-device list

        Get the list of sub-devices in the qcloud
        Args: None
        Returns:
            success: sub-device list
        """
        self.__assert(product_id)
        self.__assert(device_name)

        gateway_topic_pub = self._topic.gateway_topic_pub
        return self.__gateway.gateway_get_subdev_bind_list(gateway_topic_pub, 0, product_id, device_name)
    
    def gatewaySubdevSubscribe(self, topic):
        """Subscribe sub-device topic

        Subscribe sub-device topic
        Args: sub-device topic
        Returns:
            success: zero and subscribe mid
        """
        return self.__gateway.gateway_subdev_subscribe(topic)

    def gatewayInit(self):
        """Gateway initialization

        Gateway initialization
        Args: None
        Returns:
            success: zero and publish mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")

        self.__gateway = Gateway(self.__host, self.__device_info.product_id, self.__device_info.device_name,
                                    self.__device_info.device_secret, websocket=self.__useWebsocket,
                                    tls=self.__tls)
        json_data = self.__device_info.json_data
        gateway_topic_sub = self._topic.gateway_topic_sub

        return self.__gateway.gateway_init(gateway_topic_sub, 0, json_data)

    def rrpcInit(self, productId, deviceName, callback):
        """RRPC initialization

        RRPC initialization
        Args:
            productId: product id
            deviceName: device name
            callback: user received message callback
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")

        client = productId + deviceName
        rrpc = Rrpc(self.__host, productId, deviceName,
                        "", websocket=self.__useWebsocket,
                        tls=self.__tls, logger=self._logger)

        rc, mid = rrpc.rrpc_init(callback)
        if rc != 0:
            self._logger.error("[rrpc] subscribe error:rc:%d" % (rc))
        else:
            self.__rrpc_map[client] = rrpc
        return rc, mid

    def rrpcReply(self, productId, deviceName, reply, length):
        """RRPC reply

        Reply rrpc request
        Args:
            productId: product id
            deviceName: device name
            reply: reply message
            length: reply message length
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__rrpc_map.keys()
                or self.__rrpc_map[client] is None):
            self._logger.error("[rrpc] not found rrpc handle for client:%s" % (client))
            return None

        rrpc = self.__rrpc_map[client]
        rc, mid = rrpc.rrpc_reply(reply)
        if rc != 0:
            self._logger.error("[rrpc] publish error:rc:%d" % (rc))
        return rc, mid
    
    def broadcastInit(self, productId, deviceName, callback):
        """Broadcast initialization

        Broadcast initialization
        Args:
            productId: product id
            deviceName: device name
            callback: user received message callback
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")

        client = productId + deviceName
        broadcast = Broadcast(self.__host, productId, deviceName,
                                "", websocket=self.__useWebsocket,
                                tls=self.__tls, logger=self._logger)

        rc, mid = broadcast.broadcast_init(callback)
        if rc != 0:
            self._logger.error("[broadcast] publish error:rc:%d" % (rc))
        else:
            self.__broadcast_map[client] = broadcast
        return rc, mid

    def shadowInit(self, productId, deviceName, callback):
        """Shadow initialization

        Shadow initialization
        Args:
            productId: product id
            deviceName: device name
            callback: user received message callback
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")

        client = productId + deviceName
        shadow = Shadow(self.__host, productId, deviceName,
                            "", websocket=self.__useWebsocket,
                            tls=self.__tls, logger=self._logger)

        rc, mid = shadow.shadow_init(callback)
        if rc != 0:
            self._logger.error("[shadow] publish error:rc:%d" % (rc))
        else:
            self.__shadow_map[client] = shadow
        return rc, mid

    def getShadow(self, productId, deviceName):
        """Get shadow cache

        Get cloud shadow cache
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__shadow_map.keys()
                or self.__shadow_map[client] is None):
            self._logger.error("[shadow] not found shadow handle for client:%s" % (client))
            return None

        shadow = self.__shadow_map[client]
        rc, mid = shadow.get_shadow(productId)
        if rc != 0:
            self._logger.error("[shadow] publish error:rc:%d" % (rc))
        return rc, mid

    def shadowJsonConstructDesireAllNull(self, productId, deviceName):
        """Construct json context

        Construct json context to reset cloud cache
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: json context
            fail: None
        """
        client = productId + deviceName
        if (client not in self.__shadow_map.keys()
                or self.__shadow_map[client] is None):
            self._logger.error("[shadow] not found shadow handle for client:%s" % (client))
            return None

        shadow = self.__shadow_map[client]
        return shadow.shadow_json_construct_desire_null(productId)

    def shadowUpdate(self, productId, deviceName, shadow_docs, length):
        """Update shadow

        Update cloud shadow cache
        Args:
            productId: product id
            deviceName: device name
            shadow_docs: update message
            length: message length
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__shadow_map.keys()
                or self.__shadow_map[client] is None):
            self._logger.error("[shadow] not found shadow handle for client:%s" % (client))
            return None

        self._logger.debug("[shadow update] %s" % (shadow_docs))
        shadow = self.__shadow_map[client]
        rc, mid = shadow.shadow_update(shadow_docs)
        if rc != 0:
            self._logger.error("topic_publish error:rc:%d" % (rc))
        return rc, mid

    def shadowJsonConstructReport(self, productId, deviceName, *args):
        """Construct json context

        Construct report json context
        Args:
            productId: product id
            deviceName: device name
            *args: variable parameters, type is device_property()
        Returns:
            success: json context
            fail: None
        """
        client = productId + deviceName
        if (client not in self.__shadow_map.keys()
                or self.__shadow_map[client] is None):
            self._logger.error("[shadow] not found shadow handle for client:%s" % (client))
            return None

        shadow = self.__shadow_map[client]
        return shadow.shadow_json_construct_report(productId, *args)

    def otaInit(self, productId, deviceName, callback):
        """Ota initialization

        Ota initialization
        Args:
            productId: product id
            deviceName: device name
            callback: user received message callback
        Returns:
            success: zero and subscribe mid
        """
        if self.__hub_state is not self.HubState.CONNECTED:
            raise self.StateError("current state is not CONNECTED")

        ota = Ota(self.__host, self.__device_info.product_id,
                            self.__device_info.device_name, self.__device_info.device_secret,
                            websocket=self.__useWebsocket, tls=self.__tls, logger=self._logger)
        topic_sub = self._topic.ota_update_topic_sub
        rc, mid = ota.ota_init(productId, deviceName, callback)
        if rc != 0:
            self._logger.error("[ota] subscribe error:rc:%d,topic:%s" % (rc, topic_sub))
            return rc, mid

        """
        等待订阅成功
        """
        cnt = 0
        while cnt < 10:
            if mid in self.__subscribe_res:
                # 收到该mid回调,且其qos>=0说明订阅完成,qos=0需另做判断
                if self.__subscribe_res[mid] >= 1:
                    break
            time.sleep(0.2)
            cnt += 1
            pass
        if cnt >= 10:
            return -1, mid

        ota.ota_manager_init()

        client = productId + deviceName
        self.__ota_map[client] = ota

        return rc, mid

    # 是否应将ota句柄传入(支持多个下载进程?)
    def otaIsFetching(self, productId, deviceName):
        """Is downloading

        Is downloading
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: True
            fail: False
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return False

        ota = self.__ota_map[client]
        return ota.ota_is_fetching()

    def otaIsFetchFinished(self, productId, deviceName):
        """Is download finished

        Is download finished
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: True
            fail: False
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return False

        ota = self.__ota_map[client]
        return ota.ota_is_fetch_finished()

    def otaReportUpgradeSuccess(self, productId, deviceName, version):
        """Report success message

        Report upgrade success message to qcloud
        Args:
            productId: product id
            deviceName: device name
            version: firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1, -1

        ota = self.__ota_map[client]
        rc, mid = ota.ota_report_upgrade_success(version)
        if rc != 0:
            self._logger.error("ota report upgrade(success) fail")
        return rc, mid

    def otaReportUpgradeFail(self, productId, deviceName, version):
        """Report fail message

        Report upgrade fail message to qcloud
        Args:
            productId: product id
            deviceName: device name
            version: firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1, -1

        ota = self.__ota_map[client]
        rc, mid = ota.ota_report_upgrade_fail(version)
        if rc != 0:
            self._logger.error("ota report upgrade(fail) fail")
        return rc, mid

    def otaIoctlNumber(self, productId, deviceName, cmdType):
        """User interaction

        User interaction with SDK to get a number, like linux kernel ioctl
        Args:
            productId: product id
            deviceName: device name
            cmdType: interaction command
        Returns:
            success: the number you want and 'success' message
            fail: negative number and error message
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1, "no handle"

        ota = self.__ota_map[client]
        return ota.ota_ioctl_number(cmdType)

    def otaIoctlString(self, productId, deviceName, cmdType, length):
        """User interaction

        User interaction with SDK to get a string, like linux kernel ioctl
        Args:
            productId: product id
            deviceName: device name
            cmdType: interaction command
            length: command length
        Returns:
            success: the string you want and 'success' message
            fail: negative number and error message
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return "null", "no handle"

        ota = self.__ota_map[client]
        return ota.ota_ioctl_string(cmdType, length)

    def otaResetMd5(self, productId, deviceName):
        """Reset md5 value

        Reset md5 value
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: zero
            fail: negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1

        ota = self.__ota_map[client]
        return ota.ota_reset_md5()

    def otaMd5Update(self, productId, deviceName,buf):
        """Update md5 value

        Calculate new message md5 and update old
        Args:
            productId: product id
            deviceName: device name
            buf: new message
        Returns:
            success: zero
            fail: negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1

        ota = self.__ota_map[client]
        return ota.ota_md5_update(buf)

    def __ota_http_deinit(self, productId, deviceName, http):
        self._logger.debug("http deinit do nothing...")

    def httpInit(self, productId, deviceName, host, url, offset, size, timeoutSec):
        """Http initialization

        Http initialization
        Args:
            productId: product id
            deviceName: device name
            host: http server host
            url: http url
            offset: http parameter 'Range' minimum
            size: http parameter 'Range' max
            timeoutSec: http overtime time
        Returns:
            success: zero
            fail: negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1

        ota = self.__ota_map[client]
        return ota.http_init(host, url, offset, size, timeoutSec)

    def httpFetch(self, productId, deviceName, buf_len):
        """Http download

        Http download
        Args:
            productId: product id
            deviceName: device name
            buf_len: download max length
        Returns:
            success: downloaded content and length
            fail: None and negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return None, -1

        ota = self.__ota_map[client]
        return ota.http_fetch(buf_len)

    def otaReportVersion(self, productId, deviceName, version):
        """Report version

        Report local firmware version
        Args:
            productId: product id
            deviceName: device name
            version: local firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1, -1

        ota = self.__ota_map[client]
        rc, mid = ota.ota_report_version(version)
        if rc != 0:
            self._logger.error("[ota] report version fail")
        return rc, mid

    def otaDownloadStart(self, productId, deviceName, offset, size):
        """Start download

        Start download
        Args:
            productId: product id
            deviceName: device name
            offset: download offset
            size: download size
        Returns:
            success: zero
            fail: negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return -1

        ota = self.__ota_map[client]
        return ota.ota_download_start(offset, size)

    def otaFetchYield(self, productId, deviceName, buf_len):
        """Http download

        Perform an http download
        Args:
            productId: product id
            deviceName: device name
            buf_len: download max length
        Returns:
            success: downloaded content and length
            fail: None and negative number
        """
        client = productId + deviceName
        if (client not in self.__ota_map.keys()
                or self.__ota_map[client] is None):
            self._logger.error("[ota] not found ota handle for client:%s" % (client))
            return None, -1

        ota = self.__ota_map[client]
        return ota.ota_fetch_yield(buf_len)

    def logInit(self, level, filePath=None, maxBytes=0, backupCount=0, enable=True):
        """Log initialization

        Log initialization
        Args:
            level: log level, type is class LoggerLevel()
            enable: enable switch
        Returns:
            success: logger handle
            fail: None
        """
        if self.__protocol is None:
            return None

        logger_level = 0
        provider = LoggerProvider()
        logger = provider.logger

        if enable is False:
            logger.disable_logger()
        else:
            if level.value == self.LoggerLevel.INFO.value:
                logger_level = logging.INFO
            elif level.value == self.LoggerLevel.DEBUG.value:
                logger_level = logging.DEBUG
            elif level.value == self.LoggerLevel.WARNING.value:
                logger_level = logging.WARNING
            elif level.value == self.LoggerLevel.ERROR.value:
                logger_level = logging.ERROR
            else:
                logger_level = logging.INFO

            logger.set_level(logger_level)
            logger.enable_logger()

            if (filePath is None or 
                maxBytes == 0 or backupCount == 0):
                self._logger.error("please set logger parameter:\n"
                                    "\tfilePath: log file path.\n"
                                    "\tmaxBytes: bytes of one file.\n"
                                    "\tbackupCount: file number.")
                return None

            logger.create_file(filePath, maxBytes, backupCount)
            self.__protocol.enable_logger(logger.get_logger())

        return logger



#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import string
import json
import threading
from hub.log.log import Log
from hub.protocol.protocol import AsyncConnClient

class SingletonType(type):
    _instance_lock = threading.Lock()
    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            with SingletonType._instance_lock:
                if not hasattr(cls, "_instance"):
                    cls._instance = super(SingletonType,cls).__call__(*args, **kwargs)
        return cls._instance

class TopicProvider(object):
    def __init__(self, product_id, device_name):
        self.__clientToken = None

        # log topic
        self.__log_topic_pub = "$log/operation/%s/%s" % (product_id, device_name)
        self.__log_topic_sub = "$log/operation/result/%s/%s" % (product_id, device_name)
        self.__is_subscribed_log_topic = False

        # system topic
        self.__sys_topic_pub = "$sys/operation/%s/%s" % (product_id, device_name)
        self.__sys_topic_sub = "$sys/operation/result/%s/%s" % (product_id, device_name)

        # gateway topic
        self.__gateway_topic_pub = "$gateway/operation/%s/%s" % (product_id, device_name)
        self.__gateway_topic_sub = "$gateway/operation/result/%s/%s" % (product_id, device_name)

        # data template topic
        self.__template_topic_pub = "$template/operation/%s/%s" % (product_id, device_name)
        self.__template_topic_sub = "$template/operation/result/%s/%s" % (product_id, device_name)

        # thing topic
        self.__thing_property_topic_pub = "$thing/up/property/%s/%s" % (product_id, device_name)
        self.__thing_property_topic_sub = "$thing/down/property/%s/%s" % (product_id, device_name)

        self.__thing_action_topic_pub = "$thing/up/action/%s/%s" % (product_id, device_name)
        self.__thing_action_topic_sub = "$thing/down/action/%s/%s" % (product_id, device_name)
        self.__thing_event_topic_pub = "$thing/up/event/%s/%s" % (product_id, device_name)
        self.__thing_event_topic_sub = "$thing/down/event/%s/%s" % (product_id, device_name)
        self.__thing_raw_topic_pub = "$thing/up/raw/%s/%s" % (product_id, device_name)
        self.__thing_raw_topic_sub = "$thing/down/raw/%s/%s" % (product_id, device_name)
        self.__thing_service_topic_pub = "$thing/up/service/%s/%s" % (product_id, device_name)
        self.__thing_service_topic_sub = "$thing/down/service/%s/%s" % (product_id, device_name)

        # ota topic
        self.__ota_report_topic_pub = "$ota/report/%s/%s" % (product_id, device_name)
        self.__ota_update_topic_sub = "$ota/update/%s/%s" % (product_id, device_name)

        # rrpc topic
        self.__rrpc_topic_pub_prefix = "$rrpc/txd/%s/%s" % (product_id, device_name)
        self.__rrpc_topic_sub_prefix = "$rrpc/rxd/%s/%s" % (product_id, device_name)

        # shadow
        self.__shadow_topic_pub = "$shadow/operation/%s/%s" % (product_id, device_name)
        self.__shadow_topic_sub = "$shadow/operation/result/%s/%s" % (product_id, device_name)

        # broadcast
        self.__broadcast_topic_sub = "$broadcast/rxd/%s/%s" % (product_id, device_name)
        pass

    @property
    def sys_topic_sub(self):
        return self.__sys_topic_sub

    @property
    def sys_topic_pub(self):
        return self.__sys_topic_pub

    @property
    def gateway_topic_sub(self):
        return self.__gateway_topic_sub

    @property
    def gateway_topic_pub(self):
        return self.__gateway_topic_pub

    @property
    def template_topic_sub(self):
        return self.__template_topic_sub

    @property
    def template_event_topic_sub(self):
        return self.__thing_event_topic_sub

    @property
    def template_event_topic_pub(self):
        return self.__thing_event_topic_pub

    @property
    def template_action_topic_sub(self):
        return self.__thing_action_topic_sub

    @property
    def template_property_topic_sub(self):
        return self.__thing_property_topic_sub

    @property
    def template_property_topic_pub(self):
        return self.__thing_property_topic_pub

    @property
    def template_action_topic_pub(self):
        return self.__thing_action_topic_pub

    @property
    def template_service_topic_sub(self):
        return self.__thing_service_topic_sub

    @property
    def template_raw_topic_sub(self):
        return self.__thing_raw_topic_sub

    @property
    def ota_update_topic_sub(self):
        return self.__ota_update_topic_sub

    @property
    def ota_report_topic_pub(self):
        return self.__ota_report_topic_pub

    @property
    def rrpc_topic_pub_prefix(self):
        return self.__rrpc_topic_pub_prefix

    @property
    def rrpc_topic_sub_prefix(self):
        return self.__rrpc_topic_sub_prefix

    @property
    def shadow_topic_pub(self):
        return self.__shadow_topic_pub

    @property
    def shadow_topic_sub(self):
        return self.__shadow_topic_sub

    @property
    def broadcast_topic_sub(self):
        return self.__broadcast_topic_sub

    @property
    def control_clientToken(self):
        return self.__clientToken

    # _on_template_downstream_topic_handler()中收到云端消息后保存client-token
    @control_clientToken.setter
    def control_clientToken(self, token):
        if token is None or len(token) == 0:
            raise ValueError('Invalid info.')
        self.__clientToken = token

class DeviceInfoProvider(object):
    def __init__(self, file_path):
        self.__file_path = file_path
        # self.__logger = logger
        self.__log_provider = LoggerProvider()
        self.__logger = self.__log_provider.logger
        self.__logger.info('device_info file {}'.format(file_path))

        self.__auth_mode = None
        self.__device_name = None
        self.__product_id = None
        self.__product_secret = None
        self.__device_secret = None
        self.__ca_file = None
        self.__cert_file = None
        self.__private_key_file = None
        self.__region = None
        if self.__logger is not None:
            self.__logger.info('device_info file {}'.format(file_path))
        with open(file_path, 'r', encoding='utf-8') as f:
            self.__json_data = json.loads(f.read())
            self.__auth_mode = self.__json_data['auth_mode']
            self.__device_name = self.__json_data['deviceName']
            self.__product_id = self.__json_data['productId']
            self.__product_secret = self.__json_data['productSecret']
            self.__device_secret = self.__json_data['key_deviceinfo']['deviceSecret']
            self.__ca_file = self.__json_data['cert_deviceinfo']['devCaFile']
            self.__cert_file = self.__json_data['cert_deviceinfo']['devCertFile']
            self.__private_key_file = self.__json_data['cert_deviceinfo']['devPrivateKeyFile']
            self.__region = self.__json_data["region"]
        if self.__logger is not None:
            self.__logger.info(
                "device name: {}, product id: {}, product secret: {}, device secret: {}".
                format(self.__device_name, self.__product_id,
                        self.__product_secret, self.__device_secret))

    def update_config_file(self, psk):
        with open(self.__file_path, '+r', encoding='utf-8') as f:
            t = f.read()
            t = t.replace(self.__device_secret, psk)
            f.seek(0, 0)
            f.write(t)
            f.truncate()

            self.__device_secret = psk
        pass

    @property
    def auth_mode(self):
        return self.__auth_mode

    @property
    def device_name(self):
        return self.__device_name

    @property
    def product_id(self):
        return self.__product_id

    @property
    def product_secret(self):
        return self.__product_secret

    @property
    def device_secret(self):
        return self.__device_secret

    @property
    def ca_file(self):
        return self.__ca_file

    @property
    def cert_file(self):
        return self.__cert_file

    @property
    def private_key_file(self):
        return self.__private_key_file

    @property
    def region(self):
        return self.__region

    @property
    def json_data(self):
        return self.__json_data

class ConnClientProvider(metaclass=SingletonType):
    """
    使用单例模式构建,保证对象只有一份
    """
    def __init__(self, host, product_id, device_name, device_secret, websocket=False, tls=True, logger=None):
        self.protocol = AsyncConnClient(host, product_id, device_name, device_secret, websocket, tls, logger)

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

class LoggerProvider(metaclass=SingletonType):
    """
    使用单例模式构建,保证对象只有一份
    """
    def __init__(self):
        self.logger = Log()

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

# import os
import logging
import threading
import queue
import urllib.request
import urllib.parse
import urllib.error
import json
import hashlib
import hmac
import base64
import random
import ssl
import socket
import string
import time
import paho.mqtt.client as mqtt
from enum import Enum
from enum import IntEnum
from Crypto.Cipher import AES
from hub.hub import QcloudHub
from explorer.services.template.template import Template

class QcloudExplorer(object):

    def __init__(self, device_file, tls=True, userdata=None, domain=None, useWebsocket=False):
        self.__device_file = device_file
        self.__tls = tls
        """ 用户传参 """
        self.__userdata = userdata
        self.__domain = domain
        self.__useWebsocket = useWebsocket
        """ 存放用户注册的回调函数 """
        self.__user_callback = {}

        self.__provider = QcloudHub(device_file, userdata, tls, domain, useWebsocket)
        self.__hub = self.__provider.hub
        """
        向hub注册mqtt disconnect回调
        explorer层清理相关资源
        """
        self.__hub.register_explorer_callback("$explorer/from/disconnect", self.__on_disconnect)

        """ 用户回调注册到hub层 """
        # self.__register_hub_event_callback()

        self.__logger = self.__hub._logger

        self.__template_map = {}

        self.__topic = self.__hub._topic

        # data template
        self.__is_subscribed_property_topic = False

        # user mqtt callback
        self.__user_on_connect = None
        self.__user_on_disconnect = None
        self.__user_on_publish = None
        self.__user_on_subscribe = None
        self.__user_on_unsubscribe = None
        self.__user_on_message = None
        pass

    class ReplyPara(object):
        def __init__(self):
            self.timeout_ms = 0
            self.code = -1
            self.status_msg = None

    class LoggerLevel(Enum):
        INFO = "info"
        DEBUG = "debug"
        WARNING = "warring"
        ERROR = "error"

    def __on_disconnect(self, client, userdata, rc):
        """
        清理数据模板资源
        """
        for template in self.__template_map.values():
            if template is not None:
                template.template_reset()
        pass

    def __handle_subdev_topic(self, topic, qos, payload):
        """ 回调用户处理 """
        if self.__user_callback[topic] is not None:
            self.__user_callback[topic](topic, qos, payload, self.__userdata)
        else:
            self.__logger.error("no callback for topic %s" % topic)

    def __handle_template(self, topic, qos, payload, userdate):
        pos = topic.rfind("/")
        device_name = topic[pos + 1:len(topic)]

        topic_split = topic[0:pos]
        pos = topic_split.rfind("/")
        product_id = topic_split[pos + 1:len(topic_split)]

        client = product_id + device_name
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return None

        template = self.__template_map[client]
        template.handle_template(topic, qos, payload, userdate)

        # """ 回调用户处理 """
        # if (topic in self.__user_callback.keys()
        #         and self.__user_callback[topic] is not None):
        #     self.__user_callback[topic](topic, qos, payload, self.__userdata)
        # else:
        #     self.__logger.error("no callback for topic %s" % topic)
        pass

    def setReconnectInterval(self, max_sec, min_sec):
        """Set mqtt reconnect interval

        Set mqtt reconnect interval
        Args:
            max_sec: reconnect max time
            min_sec: reconnect min time
        Returns:
            success: default
            fail: default
        """
        self.__hub.setReconnectInterval(max_sec, min_sec)

    def setMessageTimout(self, timeout):
        """Set message overtime time

        Set message overtime time
        Args:
            timeout: mqtt keepalive value
        Returns:
            success: default
            fail: default
        """
        self.__hub.setMessageTimout(timeout)

    def setKeepaliveInterval(self, interval):
        """Set mqtt keepalive interval

        Set mqtt keepalive interval
        Args:
            interval: mqtt keepalive interval
        Returns:
            success: default
            fail: default
        """
        self.__hub.setKeepaliveInterval(interval)

    def subscribe(self, topic, qos):
        """Subscribe topic

        Subscribe topic
        Args:
            topic: topic
            qos: mqtt qos
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        return self.__hub.subscribe(topic, qos)

    def unsubscribe(self, topic):
        """Unsubscribe topic

        Unsubscribe topic what is subscribed
        Args:
            topic: topic
        Returns:
            success: zero and unsubscribe mid
            fail: negative number and unsubscribe mid
        """
        return self.__hub.unsubscribe(topic)

    def publish(self, topic, payload, qos):
        """Publish message

        Publish message
        Args:
            topic: topic
            payload: publish message
            qos: mqtt qos
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        return self.__hub.publish(topic, payload, qos)

    def isMqttConnected(self):
        """Is mqtt connected

        Is mqtt connected
        Args: None
        Returns:
            success: True/False
        """
        return self.__hub.isMqttConnected()

    def getConnectState(self):
        """Get connect state

        Get device current connect state
        Args: None
        Returns:
            success: connect state
        """
        return self.__hub.getConnectState()

    def getNtpAccurateTime(self):
        """Get NTP time

        Get NTP time
        Args: None
        Returns:
            success: thread start result
            fail: thread start result
        """
        return self.__hub.getNtpAccurateTime()

    # start thread to connect and loop
    def connect(self):
        """Connect

        Device connect
        Args: None
        Returns:
            success: thread start result
            fail: thread start result
        """
        return self.__hub.connect()

    def disconnect(self):
        """Disconnect

        Device disconnect
        Args: None
        Returns:
            success: default
            fail: default
        """
        self.__hub.disconnect()

    def registerMqttCallback(self, on_connect, on_disconnect,
                            on_message, on_publish,
                            on_subscribe, on_unsubscribe):
        """Register user mqtt callback

        Register user mqtt callback for mqtt
        Args:
            on_connect: mqtt connect callback
            on_disconnect: mqtt disconnect callback
            on_message: mqtt message callback
            on_publish: mqtt publish callback
            on_subscribe: mqtt subscribe callback
            on_unsubscribe: mqtt unsubscribe callback
        Returns:
            success: default
            fail: default
        """
        self.__hub.registerMqttCallback(on_connect, on_disconnect,
                                        on_message, on_publish,
                                        on_subscribe, on_unsubscribe)

    def registerUserCallback(self, topic, callback):
        """Register user callback

        Register user callback for a topic
        Args:
            topic: topic
            callback: user callback
        Returns:
            success: default
            fail: default
        """
        self.__user_callback[topic] = callback

    def getEventsList(self, productId, deviceName):
        """Get template event list

        Get template event list from configuration file
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: template event list
            fail: empty list
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return None

        template = self.__template_map[client]
        return template.get_events_list()

    def getActionList(self, productId, deviceName):
        """Get template action list

        Get template action list from configuration file
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: template action list
            fail: empty list
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return None

        template = self.__template_map[client]
        return template.get_action_list()

    def getPropertyList(self, productId, deviceName):
        """Get template property list

        Get template property list from configuration file
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: template property list
            fail: empty list
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return None

        template = self.__template_map[client]
        return template.get_property_list()

    def getProductID(self):
        """Get product id

        Get product id
        Args: None
        Returns:
            success: product id
            fail: None
        """
        return self.__hub.getProductID()

    def getDeviceName(self):
        """Get device name

        Get device name
        Args: None
        Returns:
            success: device name
            fail: None
        """
        return self.__hub.getDeviceName()

    def templateSetup(self, productId, deviceName, config_file=None):
        """Parse json configuration file

        Parse json configuration file
        Args:
            productId: product id
            deviceName: device name
            config_file: configuration file path
        Returns:
            success: zero
            fail: negative number
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        return template.template_setup(config_file)

    # 暂定传入json格式
    def templateEventPost(self, productId, deviceName, message):
        """Report event/events infomation

        Report device event/events infomation
        Args:
            productId: product id
            deviceName: device name
            message: device event/events infomation
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_event_post(productId, message)
        if rc != 0:
            self.__logger.error("[template] publish error:rc:%d" % (rc))
        return rc, mid

    def templateJsonConstructReportArray(self, productId, deviceName, payload):
        """Construct json array

        Construct json array
        Args:
            productId: product id
            deviceName: device name
            payload: report message, json type
        Returns:
            success: json message
            fail: None
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        return template.template_json_construct_report_array(productId, payload)

    def templateReportSysInfo(self, productId, deviceName, sysInfo):
        """Report system infomation

        Report device system infomation
        Args:
            productId: product id
            deviceName: device name
            sysInfo: device system infomation
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_report_sys_info(productId, sysInfo)
        if rc != 0:
            self.__logger.error("[template] publish error:rc:%d" % (rc))
        return rc, mid

    def templateControlReply(self, productId, deviceName, replyPara):
        """Report control reply

        Report control reply after recvive control message
        Args:
            productId: product id
            deviceName: device name
            replyPara: description infomation, type is class ReplyPara()
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_control_reply(replyPara)
        if rc != 0:
            self.__logger.error("[template] publish error:rc:%d" % (rc))
        return rc, mid

    def templateActionReply(self, productId, deviceName, clientToken, response, replyPara):
        """Report action reply

        Report action reply after recvive action message
        Args:
            productId: product id
            deviceName: device name
            clientToken: client token
            response: report message
            replyPara: other description infomation, type is class ReplyPara()
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_action_reply(clientToken, response, replyPara)
        if rc != 0:
            self.__logger.error("[template] publish error:rc:%d" % (rc))
        return rc, mid

    # 回调中处理IOT_Template_ClearControl
    def templateGetStatus(self, productId, deviceName):
        """Get status

        Get device status
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        return template.template_get_status(productId)

    def templateReport(self, productId, deviceName, message):
        """Template message report

        Report message to cloud
        Args:
            productId: product id
            deviceName: device name
            message: report message
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        # 判断下行topic是否订阅
        if self.__is_subscribed_property_topic is False:
            self.__logger.error("Template is not initialization, please do it!")
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_report(message)
        if rc != 0:
            self.__logger.error("template publish error:rc:%d" % (rc))

        return rc, mid

    def templateInit(self, productId, deviceName,
                        propertyCb, actionCb, eventCb, serviceCb):
        """Template initialization

        Template initialization
        Args:
            productId: product id
            deviceName: device name
            propertyCb: user received property message callback
            actionCb: user received action message callback
            eventCb: user received event message callback
            serviceCb: user received service message callback
        Returns:
            success: zero and subscribe mid
            fail: negative number and subscribe mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        """
        构造对应client的template对象并加入字典
        """
        client = productId + deviceName
        template = Template(self.__device_file, self.__tls, productId, deviceName,
                                self.__userdata, self.__domain, self.__useWebsocket, self.__logger)

        """
        注册用户数据模板topic(property/action/event)对应回调
        注册后用户不用再关注相关topic
        """
        rc, mid = template.template_init(self.__handle_template,
                                            propertyCb, actionCb, eventCb, serviceCb)
        if rc != 0:
            self.__logger.error("[template] subscribe error:rc:%d" % (rc))
            return rc, mid

        # save template client
        self.__template_map[client] = template
        self.__is_subscribed_property_topic = True
        return rc, mid

    def clearControl(self, productId, deviceName):
        """Clear control

        Clear control message
        Args:
            productId: device product_id
            deviceName: device device_name
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]

        # clientToken = self.__topic.control_clientToken
        rc, mid = template.template_clear_control()
        if rc != 0:
            self.__logger.error("[template] publish error:rc:%d" % (rc))
        return rc, mid

    def templateDeinit(self, productId, deviceName):
        """Template destroy

        Template destroy
        Args:
            productId: device product_id
            deviceName: device device_name
        Returns:
            success: zero and unsubscribe mid
            fail: negative number and unsubscribe mid
        """
        if self.__hub.getConnectState() is not self.__hub.HubState.CONNECTED:
            raise self.__hub.StateError("current state is not CONNECTED")

        client = productId + deviceName
        if (client not in self.__template_map.keys()
                or self.__template_map[client] is None):
            self.__logger.error("[template] not found template handle for client:%s" % (client))
            return -1, -1

        template = self.__template_map[client]
        rc, mid = template.template_deinit()
        if rc != 0:
            self.__logger.error("[template] unsubscribe error:rc:%d" % (rc))
        else:
            self.__is_subscribed_property_topic = False

        self.__template_map.pop(client)
        return rc, mid

    def dynregDevice(self, timeout=10):
        """Dynamic register

        Get the device secret from the Cloud
        Args:
            timeout: request timeout
        Returns:
            success: return zero and device secret
            fail: -1 and error message
        """
        return self.__hub.dynregDevice(timeout)

    # gateway
    def isSubdevStatusOnline(self, sub_productId, sub_devName):
        """Sub-device status

        Determine if the device is online
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: if device is online
        """
        return self.__hub.isSubdevStatusOnline(sub_productId, sub_devName)

    def updateSubdevStatus(self, sub_productId, sub_devName, status):
        """Update device status

        Update sub-device local status
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            status: new status
        Returns:
            success: None
        """
        return self.__hub.updateSubdevStatus(sub_productId, sub_devName, status)

    def gatewaySubdevGetConfigList(self):
        """Get sub-device list

        Get the list of sub-devices in the configuration file
        Args: None
        Returns:
            success: sub-device list
        """
        return self.__hub.gatewaySubdevGetConfigList()

    def gatewaySubdevSubscribe(self, topic):
        """Subscribe sub-device topic

        Subscribe sub-device topic
        Args: sub-device topic
        Returns:
            success: zero and subscribe mid
        """
        self.__hub.register_explorer_callback(topic, self.__handle_subdev_topic)
        return self.__hub.gatewaySubdevSubscribe(topic)

    def gatewaySubdevOnline(self, sub_productId, sub_devName):
        """Make sub-device online

        Make sub-device online
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: zero and publish mid
        """
        return self.__hub.gatewaySubdevOnline(sub_productId, sub_devName)

    def gatewaySubdevOffline(self, sub_productId, sub_devName):
        """Make sub-device offline

        Make sub-device offline
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
        Returns:
            success: zero and publish mid
        """
        return self.__hub.gatewaySubdevOffline(sub_productId, sub_devName)

    def gatewaySubdevBind(self, sub_productId, sub_devName, sub_secret):
        """Bind sub-device

        Gateway device bind sub-device
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            sub_secret: sub-device secret
        Returns:
            success: zero and publish mid
        """
        return self.__hub.gatewaySubdevBind(sub_productId, sub_devName, sub_secret)

    def gatewaySubdevUnbind(self, sub_productId, sub_devName):
        """Unbind sub-device

        Gateway device unbind sub-device
        Args:
            sub_productId: sub-device product_id
            sub_devName: sub-device device_name
            sub_secret: sub-device secret
        Returns:
            success: zero and publish mid
        """
        return self.__hub.gatewaySubdevUnbind(sub_productId, sub_devName)

    def gatewayInit(self):
        """Gateway initialization

        Gateway initialization
        Args: None
        Returns:
            success: zero and subscribe mid
        """
        return self.__hub.gatewayInit()

    # ota
    def otaInit(self, productId, deviceName, callback):
        """Ota initialization

        Ota initialization
        Args:
            productId: product id
            deviceName: device name
            callback: user received message callback
        Returns:
            success: zero and subscribe mid
        """
        return self.__hub.otaInit(productId, deviceName, callback)

    def otaIsFetching(self, productId, deviceName):
        """Is downloading

        Is downloading
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: True
            fail: False
        """
        return self.__hub.otaIsFetching(productId, deviceName)

    def otaIsFetchFinished(self, productId, deviceName):
        """Is download finished

        Is download finished
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: True
            fail: False
        """
        return self.__hub.otaIsFetchFinished(productId, deviceName)

    def otaReportUpgradeSuccess(self, productId, deviceName, version):
        """Report success message

        Report upgrade success message to qcloud
        Args:
            productId: product id
            deviceName: device name
            version: firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        return self.__hub.otaReportUpgradeSuccess(productId, deviceName, version)

    def otaReportUpgradeFail(self, productId, deviceName, version):
        """Report fail message

        Report upgrade fail message to qcloud
        Args:
            productId: product id
            deviceName: device name
            version: firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        return self.__hub.otaReportUpgradeFail(productId, deviceName, version)

    def otaIoctlNumber(self, productId, deviceName, cmdType):
        """User interaction

        User interaction with SDK to get a number, like linux kernel ioctl
        Args:
            productId: product id
            deviceName: device name
            cmdType: interaction command
        Returns:
            success: the number you want and 'success' message
            fail: negative number and error message
        """
        return self.__hub.otaIoctlNumber(productId, deviceName, cmdType)

    def otaIoctlString(self, productId, deviceName, cmdType, length):
        """User interaction

        User interaction with SDK to get string, like linux kernel ioctl
        Args:
            productId: product id
            deviceName: device name
            cmdType: interaction command
            length: command length
        Returns:
            success: the string you want and 'success' message
            fail: negative number and error message
        """
        return self.__hub.otaIoctlString(productId, deviceName, cmdType, length)

    def otaResetMd5(self, productId, deviceName):
        """Reset md5 value

        Reset md5 value
        Args:
            productId: product id
            deviceName: device name
        Returns:
            success: zero
            fail: negative number
        """
        return self.__hub.otaResetMd5(productId, deviceName)

    def otaMd5Update(self, productId, deviceName, buf):
        """Update md5 value

        Calculate new message md5 and update old
        Args:
            productId: product id
            deviceName: device name
            buf: new message
        Returns:
            success: zero
            fail: negative number
        """
        return self.__hub.otaMd5Update(productId, deviceName, buf)

    def httpInit(self, productId, deviceName, host, url, offset, size, timeoutSec):
        """Http initialization

        Http initialization
        Args:
            productId: product id
            deviceName: device name
            host: http server host
            url: http url
            offset: http parameter 'Range' minimum
            size: http parameter 'Range' max
            timeoutSec: http overtime time
        Returns:
            success: zero
            fail: negative number
        """
        return self.__hub.httpInit(productId, deviceName, host, url, offset, size, timeoutSec)

    def httpFetch(self, productId, deviceName, buf_len):
        """Http download

        Http download
        Args:
            productId: product id
            deviceName: device name
            buf_len: download max length
        Returns:
            success: downloaded content and length
            fail: None and negative number
        """
        return self.__hub.httpFetch(productId, deviceName, buf_len)

    def otaReportVersion(self, productId, deviceName, version):
        """Report version

        Report local firmware version
        Args:
            productId: product id
            deviceName: device name
            version: local firmware version
        Returns:
            success: zero and publish mid
            fail: negative number and publish mid
        """
        return self.__hub.otaReportVersion(productId, deviceName, version)

    def otaDownloadStart(self, productId, deviceName, offset, size):
        """Start download

        Start download
        Args:
            productId: product id
            deviceName: device name
            offset: download offset
            size: download size
        Returns:
            success: zero
            fail: negative number
        """
        return self.__hub.otaDownloadStart(productId, deviceName, offset, size)

    def otaFetchYield(self, productId, deviceName, buf_len):
        """Http download

        Perform an http download
        Args:
            productId: product id
            deviceName: device name
            buf_len: download max length
        Returns:
            success: downloaded content and length
            fail: None and negative number
        """
        return self.__hub.otaFetchYield(productId, deviceName, buf_len)

    def logInit(self, level, filePath, maxBytes, backupCount, enable=True):
        """Log initialization

        Log initialization
        Args:
            level: log level, type is class LoggerLevel()
            enable: enable switch
        Returns:
            success: logger handle
            fail: None
        """
        return self.__hub.logInit(level, filePath, maxBytes, backupCount, enable)



#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import json
# from explorer.providers.providers import Providers
from hub.hub import QcloudHub
from hub.utils.providers import TopicProvider

class Template(object):
    def __init__(self, device_file, tls, productId, deviceName, userdata=None, domain=None, useWebsocket=False, logger=None):
        self.__logger = logger
        # self.__provider = Providers(device_file, tls)
        # self.__hub = self.__provider.hub
        self.__provider = QcloudHub(device_file, userdata, tls, domain, useWebsocket)
        self.__hub = self.__provider.hub

        self.__topic = TopicProvider(productId, deviceName)
        self._template_token_num = 0

        self.__template_events_list = []
        self.__template_action_list = []
        self.__template_property_list = []
        self.__template_setup_state = False

        """
        保存用户注册的回调
        """
        self.__user_callback = {}

        # data template reply
        self.__replyAck = -1

    # property结构
    class template_property(object):
        def __init__(self):
            self.key = None
            self.data = None
            self.data_buff_len = 0
            self.type = None

    class template_action(object):
        def __init__(self):
            self.action_id = None
            self.timestamp = 0
            self.input_num = 0
            self.output_num = 0
            self.actions_input_prop = []
            self.actions_output_prop = []

        def action_input_append(self, prop):
            self.actions_input_prop.append(prop)

        def action_output_append(self, prop):
            self.actions_output_prop.append(prop)

    # event结构(sEvent)
    class template_event(object):
        def __init__(self):
            self.event_name = None
            self.type = None
            self.timestamp = 0
            self.eventDataNum = 0
            self.events_prop = []

        def event_append(self, prop):
            self.events_prop.append(prop)

    def __assert(self, param):
        if param is None or len(param) == 0:
            raise ValueError('Invalid param.')

    def __build_empty_json(self, info_in, method_in):
        client_token = info_in + "-" + str(self._template_token_num)
        self._template_token_num += 1
        if method_in is None or len(method_in) == 0:
            json_out = {
                "clientToken": client_token
            }
        else:
            json_out = {
                "method": method_in,
                "clientToken": client_token
            }
        return json_out

    def __build_action_reply(self, clientToken, response, replyPara):
        json_out = None
        json_out = {
            "method": "action_reply",
            "code": replyPara.code,
            "clientToken": clientToken,
            "status": replyPara.status_msg,
            "response": response
        }

        return json_out

    def __build_control_reply(self, token, replyPara):
        json_out = None
        if len(replyPara.status_msg) > 0:
            json_out = {
                "code": replyPara.code,
                "clientToken": token,
                "status": replyPara.status_msg
            }
        else:
            json_out = {
                "code": replyPara.code,
                "clientToken": token
            }
        return json_out

    # 构建系统信息上报的json消息
    def __json_construct_sysinfo(self, id, info_in):
        json_token = self.__build_empty_json(id, None)
        client_token = json_token["clientToken"]
        info_out = {
            "method": "report_info",
            "clientToken": client_token,
            "params": info_in
        }

        return info_out

    def __handle_reply(self, method, payload):
        clientToken = payload["clientToken"]
        replyAck = payload["code"]
        if method == "get_status_reply":
            if replyAck == 0:
                # update client token
                self.__topic.control_clientToken = clientToken
            else:
                self.__replyAck = replyAck
                self.__logger.debug("replyAck:%d" % replyAck)
        else:
            self.__replyAck = replyAck
        pass

    def __handle_control(self, payload):
        clientToken = payload["clientToken"]
        self.__topic.control_clientToken = clientToken

    def __handle_property(self, payload):
        method = payload["method"]
        if method == "control":
            self.__handle_control(payload)
        else:
            self.__handle_reply(method, payload)

    def get_events_list(self):
        return self.__template_events_list

    def get_action_list(self):
        return self.__template_action_list

    def get_property_list(self):
        return self.__template_property_list

    def handle_template(self, topic, qos, payload, userdata):
        if topic == self.__topic.template_property_topic_sub:
            # __handle_reply回调到用户，由用户调用clearContrl()
            self.__handle_property(payload)

        """
        回调用户数据模板topic对应回调
        """
        if self.__user_callback[topic] is not None:
                self.__user_callback[topic](topic, qos, payload, userdata)
        else:
            self.__logger.error("no callback for topic %s" % topic)

    def template_reset(self):
        self._template_token_num = 0
        self.__template_events_list.clear()
        self.__template_action_list.clear()
        self.__template_property_list.clear()
        self.__template_setup_state = False
        self.__replyAck = -1

    def template_deinit(self):
        # topic_list = []
        # topic_list.append(self.__topic.template_property_topic_sub)
        # topic_list.append(self.__topic.template_event_topic_sub)
        # topic_list.append(self.__topic.template_action_topic_sub)
        # topic_list.append(self.__topic.template_service_topic_sub)
        # return self.__hub.unsubscribe(topic_list)

        self.__hub.unsubscribe(self.__topic.template_property_topic_sub)
        self.__hub.unsubscribe(self.__topic.template_event_topic_sub)
        self.__hub.unsubscribe(self.__topic.template_action_topic_sub)
        return self.__hub.unsubscribe(self.__topic.template_service_topic_sub)

    def template_init(self, callback, peopery_cb, action_cb, event_cb, service_cb):
        property_topic = self.__topic.template_property_topic_sub
        action_topic = self.__topic.template_action_topic_sub
        event_topic = self.__topic.template_event_topic_sub
        service_topic = self.__topic.template_service_topic_sub
        self.__user_callback[property_topic] = peopery_cb
        self.__user_callback[action_topic] = action_cb
        self.__user_callback[event_topic] = event_cb
        self.__user_callback[service_topic] = service_cb

        topic_list = []
        topic_list.append(property_topic)
        topic_list.append(action_topic)
        topic_list.append(event_topic)
        topic_list.append(service_topic)

        self.__hub.register_explorer_callback(topic_list, callback)

        self.__hub.subscribe(property_topic, 0)
        self.__hub.subscribe(action_topic, 0)
        self.__hub.subscribe(event_topic, 0)

        return self.__hub.subscribe(service_topic, 0)

    def template_report(self, message):
        self.__assert(message)
        return self.__hub.publish(self.__topic.template_property_topic_pub, message, 0)

    def template_get_status(self, id):
        token = self.__build_empty_json(id, "get_status")
        return self.__hub.publish(self.__topic.template_property_topic_pub, token, 0)

    def template_action_reply(self, clientToken, response, replyPara):
        self.__assert(clientToken)
        self.__assert(response)
        json_out = self.__build_action_reply(clientToken, response, replyPara)
        return self.__hub.publish(self.__topic.template_action_topic_pub, json_out, 0)

    # IOT_Template_ClearControl
    def template_clear_control(self):
        clientToken = self.__topic.control_clientToken

        message = {
            "method": "clear_control",
            "clientToken": clientToken
        }
        return self.__hub.publish(self.__topic.template_property_topic_pub, message, 0)

    def template_control_reply(self, replyPara):
        json_out = self.__build_control_reply(self.__topic.control_clientToken, replyPara)
        return self.__hub.publish(self.__topic.template_property_topic_pub, json_out, 0)

    def template_report_sys_info(self, pid, sysInfo):
        self.__assert(sysInfo)

        json_out = self.__json_construct_sysinfo(pid, sysInfo)
        return self.__hub.publish(self.__topic.template_property_topic_pub, json_out, 0)

    def template_json_construct_report_array(self, pid, payload):
        self.__assert(pid)
        self.__assert(payload)

        json_token = self.__build_empty_json(pid, None)
        client_token = json_token["clientToken"]
        json_out = {
            "method": "report",
            "clientToken": client_token,
            "params": payload
        }

        return json_out

    def template_event_post(self, pid, message):
        self.__assert(pid)
        self.__assert(message)

        json_token = self.__build_empty_json(pid, None)
        client_token = json_token["clientToken"]
        events = message["events"]

        method = "event_post"
        if len(events) > 1:
            method = "events_post"
        json_out = {
            "method": method,
            "clientToken": client_token,
            "events": events
        }
        return self.__hub.publish(self.__topic.template_event_topic_pub, json_out, 1)

    def template_setup(self, config_file=None):
        if self.__template_setup_state:
            return 0
        try:
            with open(config_file, encoding='utf-8') as f:
                cfg = json.load(f)
                index = 0
                while index < len(cfg["events"]):
                    # 解析events json
                    params = cfg["events"][index]["params"]

                    p_event = self.template_event()

                    p_event.event_name = cfg["events"][index]["id"]
                    p_event.type = cfg["events"][index]["type"]
                    p_event.timestamp = 0
                    p_event.eventDataNum = len(params)

                    i = 0
                    while i < p_event.eventDataNum:
                        event_prop = self.template_property()
                        event_prop.key = params[i]["id"]
                        event_prop.type = params[i]["define"]["type"]

                        if event_prop.type == "int" or event_prop.type == "bool":
                            event_prop.data = 0
                        elif event_prop.type == "float":
                            event_prop.data = 0.0
                        elif event_prop.type == "string":
                            event_prop.data = ''
                        else:
                            self.__logger.error("type not support")
                            event_prop.data = None

                        p_event.event_append(event_prop)
                        i += 1
                    pass

                    self.__template_events_list.append(p_event)
                    index += 1

                index = 0
                while index < len(cfg["actions"]):
                    # 解析actions json
                    inputs = cfg["actions"][index]["input"]
                    outputs = cfg["actions"][index]["output"]

                    p_action = self.template_action()
                    p_action.action_id = cfg["actions"][index]["id"]
                    p_action.input_num = len(inputs)
                    p_action.output_num = len(outputs)
                    p_action.timestamp = 0

                    i = 0
                    while i < p_action.input_num:
                        action_prop = self.template_property()
                        action_prop.key = inputs[i]["id"]
                        action_prop.type = inputs[i]["define"]["type"]

                        if action_prop.type == "int" or action_prop.type == "bool":
                            action_prop.data = 0
                        elif action_prop.type == "float":
                            action_prop.data = 0.0
                        elif action_prop.type == "string":
                            action_prop.data = ''
                        else:
                            self.__logger.error("type not support")
                            action_prop.data = None
                        p_action.action_input_append(action_prop)
                        i += 1
                    pass

                    i = 0
                    while i < p_action.output_num:
                        action_prop = self.template_property()
                        action_prop.key = outputs[i]["id"]
                        action_prop.type = outputs[i]["define"]["type"]

                        if action_prop.type == "int" or action_prop.type == "bool":
                            action_prop.data = 0
                        elif action_prop.type == "float":
                            action_prop.data = 0.0
                        elif action_prop.type == "string":
                            action_prop.data = ''
                        else:
                            self.__logger.error("type not support")
                            action_prop.data = None
                        p_action.action_output_append(action_prop)
                        i += 1
                    pass

                    self.__template_action_list.append(p_action)
                    index += 1
                pass

                index = 0
                while index < len(cfg["properties"]):
                    # 解析properties json
                    p_prop = self.template_property()
                    p_prop.key = cfg["properties"][index]["id"]
                    p_prop.type = cfg["properties"][index]["define"]["type"]

                    if p_prop.type == "int" or p_prop.type == "bool" or p_prop.type == "enum":
                        p_prop.data = 0
                    elif p_prop.type == "float":
                        p_prop.data = 0.0
                    elif p_prop.type == "string":
                        p_prop.data = ''
                    else:
                        self.__logger.error("type not support")
                        p_prop.data = None

                    self.__template_property_list.append(p_prop)
                    index += 1
                pass

        except Exception as e:
            self.__logger.error("config file open error:" + str(e))
            return 2
        self.__template_setup_state = True
        return 0
     


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import json
from hub.utils.providers import ConnClientProvider
from hub.utils.providers import TopicProvider

class Rrpc(object):
    def __init__(self, host, product_id, device_name, device_secret,
                    websocket=False, tls=True, logger=None):
        self.__provider = ConnClientProvider(host, product_id, device_name, device_secret,
                                                websocket=websocket, tls=tls, logger=logger)
        self.__protocol = self.__provider.protocol
        self.__topic = TopicProvider(product_id, device_name)
        self.__logger = logger
        self.__process_id = None
        self.__user_callback = {}

    def __assert(self, param):
        if param is None or len(param) == 0:
            raise ValueError('Invalid param.')

    def __rrpc_get_process_id(self, topic):
        pos = topic.rfind("/")
        if pos > 0:
            self.__process_id = topic[pos + 1:len(topic)]
            return 0
        else:
            self.__logger.error("cannot found process id from topic:%s" % topic)
            return -1

    def handle_rrpc(self, topic, qos, payload, userdata):
        self.__rrpc_get_process_id(topic)

        pos = topic.rfind("/")
        topic_split = topic[0:pos]
        if self.__user_callback[topic_split] is not None:
                self.__user_callback[topic_split](topic_split, qos, payload, userdata)
        else:
            self.__logger.error("no callback for topic_split %s" % topic_split)

    def rrpc_init(self, rrpc_cb):
        topic = self.__topic.rrpc_topic_sub_prefix + "/+"
        self.__user_callback[self.__topic.rrpc_topic_sub_prefix] = rrpc_cb
        

        return self.__protocol.subscribe(topic, 0)

    def rrpc_reply(self, reply):
        self.__assert(self.__process_id)
        self.__assert(reply)

        topic = self.__topic.rrpc_topic_pub_prefix + "/" + self.__process_id
        return self.__protocol.publish(topic, json.dumps(reply), 0)


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import json
import time
import urllib.request
import urllib.parse
import urllib.error
import hashlib
from enum import Enum
from enum import IntEnum
from hub.utils.codec import Codec
from hub.utils.providers import TopicProvider
from hub.utils.providers import ConnClientProvider

class Ota(object):
    def __init__(self, host, product_id, device_name, device_secret,
                    websocket=False, tls=True, logger=None):
        self.__provider = ConnClientProvider(host, product_id, device_name, device_secret,
                                                websocket=websocket, tls=tls, logger=logger)
        self.__protocol = self.__provider.protocol
        self.__logger = logger
        self.__codec = Codec()
        self.__topic = None

        self.__ota_manager = None
        self.__ota_version_len_min = 1
        self.__ota_version_len_max = 32
        self.http_manager = None
        self.__user_callback = {}

    class OtaState(Enum):
        IOT_OTAS_UNINITED = 0
        IOT_OTAS_INITED = 1
        IOT_OTAS_FETCHING = 2
        IOT_OTAS_FETCHED = 3
        IOT_OTAS_DISCONNECTED = 4

    class OtaCmdType(Enum):
        IOT_OTAG_FETCHED_SIZE = 0
        IOT_OTAG_FILE_SIZE = 1
        IOT_OTAG_MD5SUM = 2
        IOT_OTAG_VERSION = 3
        IOT_OTAG_CHECK_FIRMWARE = 4

    class OtaProgressCode(Enum):
        IOT_OTAP_BURN_FAILED = -4
        IOT_OTAP_CHECK_FALIED = -3
        IOT_OTAP_FETCH_FAILED = -2
        IOT_OTAP_GENERAL_FAILED = -1
        IOT_OTAP_FETCH_PERCENTAGE_MIN = 0
        IOT_OTAP_FETCH_PERCENTAGE_MAX = 100

    class OtaReportType(IntEnum):
        IOT_OTAR_DOWNLOAD_TIMEOUT = -1
        IOT_OTAR_FILE_NOT_EXIST = -2
        IOT_OTAR_AUTH_FAIL = -3
        IOT_OTAR_MD5_NOT_MATCH = -4
        IOT_OTAR_UPGRADE_FAIL = -5
        IOT_OTAR_NONE = 0
        IOT_OTAR_DOWNLOAD_BEGIN = 1
        IOT_OTAR_DOWNLOADING = 2
        IOT_OTAR_UPGRADE_BEGIN = 3
        IOT_OTAR_UPGRADE_SUCCESS = 4

    class ota_manage(object):
        def __init__(self):
            self.channel = None
            self.state = 0
            self.size_fetched = 0
            self.size_last_fetched = 0
            self.file_size = 0
            self.purl = None
            self.version = None
            self.md5sum = None
            self.md5 = None
            self.host = None
            self.is_https = False

            self.report_timestamp = 0

            # http连接管理
            self.http_manager = None

    class http_manage(object):
        def __init__(self):
            self.handle = None
            self.request = None
            self.header = None
            self.host = None
            self.https_context = None

            self.err_reason = None
            self.err_code = 0
            pass

    def __assert(self, param):
        if param is None or len(param) == 0:
            raise ValueError('Invalid param.')

    def __ota_publish(self, message, qos):
        topic_pub = self.__topic.ota_report_topic_pub
        rc, mid = self.__protocol.publish(topic_pub, json.dumps(message), qos)
        return rc, mid

    def __ota_info_get(self, payload):
        size = payload["file_size"]
        if size > 0:
            self.__ota_manager.file_size = size
        version = payload["version"]
        if version is not None and len(version) > 0:
            self.__ota_manager.version = version
        url = payload["url"]
        if url is not None and len(url) > 0:
            self.__ota_manager.purl = url
            pos = url.find("https://")
            last_pos = url.rfind("/")
            if pos >= 0:
                self.__ota_manager.is_https = True
                host = url[8:last_pos]
            else:
                host = url[7:last_pos]
            self.__ota_manager.host = host

        md5sum = payload["md5sum"]
        if md5sum is not None and len(md5sum) > 0:
            self.__ota_manager.md5sum = md5sum

        self.__ota_manager.state = self.OtaState.IOT_OTAS_FETCHING

    def __ota_http_deinit(self, http):
        print("__ota_http_deinit do nothing")

    def __message_splice(self, state, progress, result_code, result_msg, version, ptype):
        message = None
        code = "%d" % (result_code)
        if ptype == 1:
            message = {
                "type": "report_progress",
                "report": {
                    "progress": {
                        "state": state,
                        "percent": str(progress),
                        "result_code": code,
                        "result_msg": result_msg
                    },
                    "version": version
                }
            }
        elif ptype == 0:
            message = {
                "type": "report_progress",
                "report": {
                    "progress": {
                        "state": state,
                        "result_code": code,
                        "result_msg": result_msg
                    },
                    "version": version
                }
            }

        return message

    def __ota_gen_report_msg(self, version, progress, report_type):
        message = None
        if report_type == self.OtaReportType.IOT_OTAR_DOWNLOAD_BEGIN:
            message = self.__message_splice("downloading", 0, 0, "", version, 1)
        elif report_type == self.OtaReportType.IOT_OTAR_DOWNLOADING:
            message = self.__message_splice("downloading", progress, 0, "", version, 1)
        elif ((report_type == self.OtaReportType.IOT_OTAR_DOWNLOAD_TIMEOUT)
                or (report_type == self.OtaReportType.IOT_OTAR_FILE_NOT_EXIST)
                or (report_type == self.OtaReportType.IOT_OTAR_MD5_NOT_MATCH)
                or (report_type == self.OtaReportType.IOT_OTAR_AUTH_FAIL)
                or (report_type == self.OtaReportType.IOT_OTAR_UPGRADE_FAIL)):
            message = self.__message_splice("fail", progress, report_type, "time_out", version, 0)
        elif report_type == self.OtaReportType.IOT_OTAR_UPGRADE_BEGIN:
            message = self.__message_splice("burning", progress, 0, "", version, 0)
        elif report_type == self.OtaReportType.IOT_OTAR_UPGRADE_SUCCESS:
            message = self.__message_splice("done", progress, 0, "", version, 0)
        else:
            self.__logger.error("not support report_type:%d" % report_type)
            message = None

        return message

    def _ota_report_upgrade_result(self, version, report_type):
        if self.__ota_manager.state == self.OtaState.IOT_OTAS_UNINITED:
            raise ValueError('ota handle is uninitialized')
        message = self.__ota_gen_report_msg(version, 1, report_type)
        if message is not None:
            return self.__ota_publish(message, 1)
        else:
            self.__logger.error("message is none")
            return -1, -1

    def _ota_report_progress(self, progress, version, report_type):
        if self.__ota_manager.state == self.OtaState.IOT_OTAS_UNINITED:
            raise ValueError('ota handle is uninitialized')
        message = self.__ota_gen_report_msg(version, progress, report_type)
        if message is not None:
            self.__logger.debug("[ota report] %s" % (message))
            return self.__ota_publish(message, 0)
        else:
            self.__logger.error("message is none")
            return -1, -1

    def ota_report_upgrade_success(self, version):
        if version is None:
            rc, mid = self._ota_report_upgrade_result(self.__ota_manager.version,
                                                     self.OtaReportType.IOT_OTAR_UPGRADE_SUCCESS)
        else:
            rc, mid = self._ota_report_upgrade_result(version, self.OtaReportType.IOT_OTAR_UPGRADE_SUCCESS)

        return rc, mid

    def ota_report_upgrade_fail(self, version):
        if version is None:
            rc, mid = self._ota_report_upgrade_result(self.__ota_manager.version,
                                                     self.OtaReportType.IOT_OTAR_UPGRADE_FAIL)
        else:
            rc, mid = self._ota_report_upgrade_result(version, self.OtaReportType.IOT_OTAR_UPGRADE_FAIL)

        return rc, mid

    def ota_ioctl_number(self, cmd_type):
        if ((self.__ota_manager.state == self.OtaState.IOT_OTAS_INITED)
                or (self.__ota_manager.state == self.OtaState.IOT_OTAS_UNINITED)):
            return -1, "state error"

        if cmd_type.value == self.OtaCmdType.IOT_OTAG_FETCHED_SIZE.value:
            return self.__ota_manager.size_fetched, "success"
        elif cmd_type.value == self.OtaCmdType.IOT_OTAG_FILE_SIZE.value:
            return self.__ota_manager.file_size, "success"
        elif cmd_type.value == self.OtaCmdType.IOT_OTAG_CHECK_FIRMWARE.value:
            if self.__ota_manager.state is not self.OtaState.IOT_OTAS_FETCHED:
                return -1, "state error"
            md5sum = self.__ota_manager.md5.hexdigest()
            if md5sum == self.__ota_manager.md5sum:
                return 0, "success"
            else:
                self._ota_report_upgrade_result(self.__ota_manager.version,
                                               self.OtaReportType.IOT_OTAR_MD5_NOT_MATCH)
                return -1, "md5 error"
            pass

        return -1, "cmd type error"

    def ota_ioctl_string(self, cmd_type, length):
        if ((self.__ota_manager.state == self.OtaState.IOT_OTAS_INITED)
                or (self.__ota_manager.state == self.OtaState.IOT_OTAS_UNINITED)):
            return "null", "state error"

        if cmd_type.value == self.OtaCmdType.IOT_OTAG_VERSION.value:
            if len(self.__ota_manager.version) > length:
                return "null", "version length error"
            else:
                return self.__ota_manager.version, "success"
        elif cmd_type.value == self.OtaCmdType.IOT_OTAG_MD5SUM.value:
            if len(self.__ota_manager.md5sum) > length:
                return "null", "md5sum length error"
            else:
                return self.__ota_manager.md5sum, "success"

        return "null", "cmd type error"

    def ota_reset_md5(self):
        self.__ota_manager.md5 = None
        self.__ota_manager.md5 = hashlib.md5()
        return 0

    def ota_md5_update(self, buf):
        if buf is None:
            self.__logger.error("buf is none")
            return -1
        if self.__ota_manager.md5 is None:
            self.__logger.error("md5 handle is uninitialized")
            return -1

        # self.__ota_manager.md5.update(buf.encode(encoding='utf-8'))
        self.__ota_manager.md5.update(buf)
        return 0

    def handle_ota(self, topic, qos, payload, userdata):
        ptype = payload["type"]
        if ptype == "report_version_rsp":
            """
            回调用户
            """
            if self.__user_callback[topic] is not None:
                self.__user_callback[topic](topic, qos, payload, userdata)
            else:
                self.__logger.error("no callback for topic %s" % topic)

        elif ptype == "update_firmware":
            self.__ota_info_get(payload)

    def ota_init(self, product_id, device_name, ota_cb):
        """
        ota资源初始化
        """
        self.__topic = TopicProvider(product_id, device_name)
        topic_sub = self.__topic.ota_update_topic_sub
        self.__user_callback[topic_sub] = ota_cb

        self.__ota_manager = self.ota_manage()
        self.__ota_manager.state = self.OtaState.IOT_OTAS_UNINITED

        return self.__protocol.subscribe(topic_sub, 1)

    def ota_manager_init(self):
        self.__ota_manager.state = self.OtaState.IOT_OTAS_INITED
        self.__ota_manager.md5 = hashlib.md5()

    # 是否应将ota句柄传入(支持多个下载进程?)
    def ota_is_fetching(self):
        return (self.__ota_manager.state == self.OtaState.IOT_OTAS_FETCHING)

    def ota_is_fetch_finished(self):
        return (self.__ota_manager.state == self.OtaState.IOT_OTAS_FETCHED)

    def http_init(self, host, url, offset, size, timeout_sec):
        range_format = "bytes=%d-%d"
        srange = range_format % (offset, size)

        header = {}
        header["Host"] = host
        header["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
        header["Accept-Encoding"] = "gzip, deflate"
        header["Range"] = srange

        self.http_manager = self.http_manage()
        self.http_manager.header = header
        self.http_manager.host = host

        if self.__ota_manager.is_https:
            # context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH, cadata=self.__iot_ca_crt)
            context = self.__codec.Ssl().create_content()
            self.http_manager.https_context = context
            try:
                self.http_manager.request = urllib.request.Request(url=url, headers=header)
                self.http_manager.handle = urllib.request.urlopen(self.http_manager.request,
                                                                  context=context,
                                                                  timeout=timeout_sec)
            except urllib.error.HTTPError as e:
                self.__logger.error("https connect error:%d" % e.code)
                self.http_manager.err_code = e.code
                return 1
            except urllib.error.URLError as e:
                self.__logger.error("https connect error:%s" % e.reason)
                self.http_manager.err_reason = e.reason
                return 1
        else:
            try:
                self.http_manager.request = urllib.request.Request(url=url, headers=header)
                self.http_manager.handle = urllib.request.urlopen(self.http_manager.request,
                                                                  timeout=timeout_sec)
            except Exception as e:
                self.__logger.error("http connect error:%s" % str(e))
                return 1
        return 0

    def http_fetch(self, buf_len):
        if self.http_manager.handle is None:
            return None, -1
        try:
            buf = self.http_manager.handle.read(buf_len)
            return buf, len(buf)
        except Exception as e:
            self.__logger.error("http read error:%s" % str(e))
            return None, -2

    def ota_report_version(self, version):
        self.__assert(version)
        if len(version) < self.__ota_version_len_min or len(version) > self.__ota_version_len_max:
            raise ValueError('Invalid version length')
        if self.__ota_manager.state == self.OtaState.IOT_OTAS_UNINITED:
            raise ValueError('ota handle is uninitialized')
        report = {
            "type": "report_version",
            "report": {
                "version": version
            }
        }
        return self.__ota_publish(report, 1)

    def ota_download_start(self, offset, size):
        if offset < 0 or size <= 0:
            raise ValueError('Invalid length.')
        if offset == 0:
            self.ota_reset_md5()
        self.__ota_http_deinit(self.__ota_manager.http_manager)
        # 断点续传初始值不为0
        self.__ota_manager.size_fetched += offset

        rc = self.http_init(self.__ota_manager.host, self.__ota_manager.purl, offset, size, 10000 / 1000)
        if rc != 0:
            if self.http_manager.err_code == 403:
                self._ota_report_upgrade_result(self.__ota_manager.version,
                                               self.OtaReportType.IOT_OTAR_AUTH_FAIL)
            elif self.http_manager.err_code == 404:
                self._ota_report_upgrade_result(self.__ota_manager.version,
                                               self.OtaReportType.IOT_OTAR_FILE_NOT_EXIST)
            elif self.http_manager.err_code == 408:
                self._ota_report_upgrade_result(self.__ota_manager.version,
                                               self.OtaReportType.IOT_OTAR_DOWNLOAD_TIMEOUT)
            else:
                # 其他错误判断(error.reason)
                self.__logger.error("http_init error:%d" % self.http_manager.err_code)

        return rc

    def ota_fetch_yield(self, buf_len):
        if self.__ota_manager.state != self.OtaState.IOT_OTAS_FETCHING:
            self.__logger.error("ota state is not fetching")
            return None, -1
        # http read
        buf, rv_len = self.http_fetch(buf_len)
        if rv_len < 0:
            if rv_len == -2:
                self._ota_report_upgrade_result(self.__ota_manager.version,
                                               self.OtaReportType.IOT_OTAR_DOWNLOAD_TIMEOUT)
            return None, -2
        else:
            if self.__ota_manager.size_fetched == 0:
                self._ota_report_progress(self.OtaProgressCode.IOT_OTAP_FETCH_PERCENTAGE_MIN,
                                         self.__ota_manager.version,
                                         self.OtaReportType.IOT_OTAR_DOWNLOAD_BEGIN)
                self.__ota_manager.report_timestamp = int(time.time())
            pass
        self.__ota_manager.size_last_fetched = rv_len
        self.__ota_manager.size_fetched += rv_len

        percent = int((self.__ota_manager.size_fetched * 100) / self.__ota_manager.file_size)
        if percent == 100:
            self._ota_report_progress(percent, self.__ota_manager.version,
                                     self.OtaReportType.IOT_OTAR_DOWNLOADING)
        else:
            timestamp = int(time.time())
            # 间隔1秒上报一次
            if (((timestamp - self.__ota_manager.report_timestamp) >= 1)
                    and (self.__ota_manager.size_last_fetched > 0)):
                self.__ota_manager.report_timestamp = timestamp
                self._ota_report_progress(percent, self.__ota_manager.version,
                                         self.OtaReportType.IOT_OTAR_DOWNLOADING)

        if self.__ota_manager.size_fetched >= self.__ota_manager.file_size:
            self.__ota_manager.state = self.OtaState.IOT_OTAS_FETCHED

        self.__ota_manager.md5.update(buf)

        return buf, rv_len


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import logging
from logging.handlers import RotatingFileHandler

class Log(object):
    def __init__(self):
        self.__logger = logging.getLogger("TecentQcloud")
        self.__format = '%(asctime)s.%(msecs)03d [%(filename)s:%(lineno)d] - %(levelname)s - %(message)s'
        logging.basicConfig(format=self.__format)
        self.__enabled = False

    def get_logger(self):
        return self.__logger

    def enable_logger(self):
        self.__enabled = True

    def disable_logger(self):
        self.__enabled = False

    def is_enabled(self):
        return self.__enabled

    def set_level(self, level):
        self.__logger.setLevel(level)

    def create_file(self, file_path, max_bytes, backup_count):
        log_handler = RotatingFileHandler(filename=file_path, maxBytes=max_bytes, backupCount=backup_count)
        log_handler.setFormatter(logging.Formatter(self.__format))
        self.__logger.addHandler(log_handler)

    def debug(self, fmt, *args):
        if self.__enabled:
            self.__logger.debug(fmt, *args)

    def warring(self, fmt, *args):
        if self.__enabled:
            self.__logger.warning(fmt, *args)

    def info(self, fmt, *args):
        if self.__enabled:
            self.__logger.info(fmt, *args)

    def error(self, fmt, *args):
        if self.__enabled:
            self.__logger.error(fmt, *args)

    def critical(self, fmt, *args):
        if self.__enabled:
            self.__logger.critical(fmt, *args)



#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import queue
import threading
from hub.utils.providers import LoggerProvider

class TaskManager(object):
    def __init__(self):
        self.__init = True

    class LoopThread(object):
        def __init__(self):
            self.__provider = LoggerProvider()
            self.__logger = self.__provider.logger
            self.__callback = None
            self.__thread = None
            self.__started = False
            self.__req_wait = threading.Event()

        def start(self, callback):
            if self.__started:
                self.__logger.info("LoopThread already")
                return 1
            else:
                self.__callback = callback
                self.__thread = threading.Thread(target=self.__thread_main)
                self.__thread.daemon = True
                self.__thread.start()
                return 0

        def __thread_main(self):
            self.__started = True
            try:
                if self.__logger is not None:
                    self.__logger.debug("LoopThread thread enter")
                if self.__callback is not None:
                    self.__callback()
                if self.__logger is not None:
                    self.__logger.debug("LoopThread thread exit")
            except Exception as e:
                self.__logger.error("LoopThread thread Exception:" + str(e))
            self.__started = False
            self.__req_wait.set()

        def stop(self):
            self.__req_wait.wait()
            self.__req_wait.clear()

    # 用户注册回调分发线程
    class EventCbThread(object):
        def __init__(self):
            self.__provider = LoggerProvider()
            self.__logger = self.__provider.logger
            self.__message_queue = queue.Queue(20)
            self.__event_callback = {}
            self.__started = False
            self.__exited = False
            self.__thread = None
            pass

        def register_event_callback(self, event, callback):
            if self.__started is False:
                if event != "req_exit":
                    self.__event_callback[event] = callback
                    return 0
                else:
                    return 1
                pass
            else:
                return 2
            pass

        def post_message(self, event, value):
            # self.__logger.debug("post_message :%r " % event)
            if self.__started and self.__exited is False:
                try:
                    self.__message_queue.put((event, value), timeout=5)
                except queue.Full as e:
                    self.__logger.error("queue full: %r" % e)
                    return False
                # self.__logger.debug("post_message success")
                return True
            self.__logger.debug("post_message fail started:%r,exited:%r" % (self.__started, self.__exited))
            return False

        def start(self):
            if self.__logger is not None:
                self.__logger.info("EventCbThread start")
            if self.__started is False:
                if self.__logger is not None:
                    self.__logger.info("EventCbThread try start")
                self.__exited = False
                self.__started = True
                self.__message_queue = queue.Queue(20)
                self.__thread = threading.Thread(target=self.__event_thread)
                self.__thread.daemon = True
                self.__thread.start()
                return 0
            return 1

        def stop(self):
            if self.__started and self.__exited is False:
                self.__exited = True
                self.__message_queue.put(("req_exit", None))

        def wait_stop(self):
            if self.__started is True:
                self.__thread.join()

        def __event_thread(self):
            if self.__logger is not None:
                self.__logger.debug("thread runnable enter")
            while True:
                event, value = self.__message_queue.get()
                if event == "req_exit":
                    break
                if self.__event_callback[event] is not None:
                    try:
                        self.__event_callback[event](value)
                    except Exception as e:
                        if self.__logger is not None:
                            self.__logger.error("thread runnable raise exception:%s" % e)
            self.__started = False
            if self.__logger is not None:
                self.__logger.debug("thread runnable exit")
            pass


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.


import sys
import socket
import string
import time
import hmac
import hashlib
import base64
import random
import ssl
import paho.mqtt.client as mqtt
from hub.utils.codec import Codec

class AsyncConnClient(object):

    def __init__(self, host, product_id, device_name, device_secret,
                    websocket=False, tls=True, logger=None):
        self.__logger = logger
        self.__tls = tls
        self.__useWebsocket = websocket
        self.__key_mode = True
        # self.__iot_ca_crt = self.__IOT_CA_CRT
        self.__mqtt_client = None
        self.__product_id = product_id
        self.__device_name = device_name
        self.__device_secret = device_secret
        self.__host = host
        self.__psk = None
        self.__certificate = self.Certificate()
        self.__codec = Codec()
        self.__set_mqtt_default_param()

        self.__init(host, product_id, device_name, device_secret)

    class Certificate:
        def __init__(self):
            self.ca_file = None
            self.cert_file = None
            self.key_file = None

    def __set_mqtt_default_param(self):
        self.__mqtt_tls_port = 8883
        self.__mqtt_tcp_port = 1883
        self.__mqtt_socket_tls_port = 443
        self.__mqtt_socket_tcp_port = 80
        self.__mqtt_protocol = "MQTTv31"
        self.__mqtt_transport = "TCP"
        self.__mqtt_secure = "TLS"
        self.__mqtt_clean_session = True
        self.__mqtt_keep_alive = 60

        self.__mqtt_auto_reconnect_min_sec = 1
        self.__mqtt_auto_reconnect_max_sec = 60
        self.__mqtt_max_queued_message = 40
        self.__mqtt_max_inflight_message = 20
        self.__mqtt_auto_reconnect_sec = 0
        self.__mqtt_request_timeout = 10

        # default MQTT/CoAP timeout value when connect/pub/sub
        self.__mqtt_command_timeout = 5
        pass

    def _generate_pwss(self, client_id, device_secret):
        self.__psk = base64.b64decode(device_secret.encode("utf-8"))

        timestamp = str(int(round(time.time() * 1000)))
        conn_id = ''.join(random.sample(string.ascii_letters + string.digits, 5))
        username = client_id + ";21010406;" + conn_id + ";" + timestamp
        sign = hmac.new(self.__psk, username.encode("utf-8"), hashlib.sha1).hexdigest()
        password = "%s;hmacsha1" % (sign)

        return username, password

    def _ssl_init(self, key_mode):
        # 密钥认证
        if key_mode is True:
            # context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH, cadata=self.__iot_ca_crt)
            self.__logger.info("connect with key...")
            context = self.__codec.Ssl().create_content()
            self.__mqtt_client.tls_set_context(context)
        else:
            self.__logger.info("connect with certificate...")
            ca = self.__certificate.ca_file
            cert = self.__certificate.cert_file
            key = self.__certificate.key_file
            self.__mqtt_client.tls_set(ca_certs=ca, certfile=cert, keyfile=key,
                                        cert_reqs=ssl.CERT_REQUIRED, tls_version=ssl.PROTOCOL_SSLv23)
        pass

    def __init(self, host, product_id, device_name, device_secret):
        client_id = product_id + device_name
        username, password = self._generate_pwss(client_id, device_secret)

        if self.__mqtt_protocol == "MQTTv311":
            mqtt_protocol_version = mqtt.MQTTv311
        elif self.__mqtt_protocol == "MQTTv31":
            mqtt_protocol_version = mqtt.MQTTv31
        if self.__useWebsocket:
            self.__mqtt_client = mqtt.Client(client_id=client_id,
                                             clean_session=self.__mqtt_clean_session,
                                             protocol=mqtt_protocol_version,
                                             transport="websockets")
        else:
            self.__mqtt_client = mqtt.Client(client_id=client_id,
                                             clean_session=self.__mqtt_clean_session,
                                             protocol=mqtt_protocol_version)

        # self._configuration_init(username, password)
        self.__mqtt_client.username_pw_set(username, password)

    def set_connect_state(self, state):
        self.__connect_state = state

    def get_connect_state(self):
        return self.__connect_state
    
    def enable_logger(self, pahoLog):
        if self.__mqtt_client is not None:
            self.__mqtt_client.enable_logger(pahoLog)

    # 设置重连时间,config_connect之前调用
    def set_reconnect_interval(self, max_sec, min_sec):
        self.__mqtt_auto_reconnect_min_sec = min_sec
        self.__mqtt_auto_reconnect_max_sec = max_sec

    # 设置消息发送超时时间,connect之前调用
    def set_message_timout(self, timeout):
        self.__mqtt_command_timeout = timeout

    # 设置保活时间,connect之前调用
    def set_keepalive_interval(self, interval):
        self.__mqtt_keep_alive = interval

    def reset_reconnect_wait(self):
        self.__mqtt_auto_reconnect_sec = 0

    def reconnect_wait(self):
        if self.__mqtt_auto_reconnect_sec == 0:
            self.__mqtt_auto_reconnect_sec = self.__mqtt_auto_reconnect_min_sec
        else:
            self.__mqtt_auto_reconnect_sec = min(self.__mqtt_auto_reconnect_sec * 2, self.__mqtt_auto_reconnect_max_sec)
            self.__mqtt_auto_reconnect_sec += random.randint(1, self.__mqtt_auto_reconnect_sec)
        time.sleep(self.__mqtt_auto_reconnect_sec)
        pass

    def register_event_callbacks(self, on_connect, on_disconnect, on_message, on_publish, on_subscribe, on_unsubscribe):
        self.__mqtt_client.on_connect = on_connect
        self.__mqtt_client.on_disconnect = on_disconnect
        self.__mqtt_client.on_message = on_message
        self.__mqtt_client.on_publish = on_publish
        self.__mqtt_client.on_subscribe = on_subscribe
        self.__mqtt_client.on_unsubscribe = on_unsubscribe

    # 配置mqtt,connect之前调用
    def config_connect(self):
        self.__mqtt_client.reconnect_delay_set(self.__mqtt_auto_reconnect_min_sec, self.__mqtt_auto_reconnect_max_sec)
        self.__mqtt_client.max_queued_messages_set(self.__mqtt_max_queued_message)
        self.__mqtt_client.max_inflight_messages_set(self.__mqtt_max_inflight_message)

    # 设置证书路径,connect之前调用
    def set_cert_file(self, ca, cert, key):
        # 认证模式置为证书方式
        self.__key_mode = False
        self.__certificate.ca_file = ca
        self.__certificate.cert_file = cert
        self.__certificate.key_file = key

    def loop(self):
        rc = mqtt.MQTT_ERR_SUCCESS
        while rc == mqtt.MQTT_ERR_SUCCESS and self.__mqtt_client is not None:
            rc = self.__mqtt_client.loop(self.__mqtt_command_timeout, 1)

    def reconnect(self):
        if self.__mqtt_client is not None:
            self.__mqtt_client.reconnect()

    def connect(self):
        mqtt_port = self.__mqtt_tls_port
        if self.__tls:
            try:
                if self.__useWebsocket:
                    mqtt_port = self.__mqtt_socket_tls_port
                self._ssl_init(self.__key_mode)
            except ssl.SSLError as e:
                self.__logger.error("ssl init error:" + str(e))
                return False
        else:
            mqtt_port = self.__mqtt_tcp_port
            if self.__useWebsocket:
                mqtt_port = self.__mqtt_socket_tcp_port
            pass
        try:
            self.__logger.debug("connect_async (%s:%d)" % (self.__host, mqtt_port))
            self.__mqtt_client.connect_async(host=self.__host, port=mqtt_port, keepalive=self.__mqtt_keep_alive)
        except Exception as e:
            self.__logger.error("mqtt connect with async error:" + str(e))
            return False
        return True

    def disconnect(self):
        if self.__mqtt_client is not None:
            self.__mqtt_client.disconnect()

    def subscribe(self, topic, qos=-1):
        rc, mid = -1, -1
        if self.__mqtt_client is None:
            return rc, mid
        
        if isinstance(topic, tuple):
            topic, qos = topic
        if isinstance(topic, str):
            if qos < 0 or qos > 1:
                raise ValueError('Invalid QoS level.')
            if topic is None or len(topic) == 0:
                raise ValueError('Invalid topic.')
            pass
            rc, mid = self.__mqtt_client.subscribe(topic, qos)
            if rc == mqtt.MQTT_ERR_SUCCESS:
                self.__logger.debug("subscribe success topic:%s" % topic)
                return 0, mid
            elif rc == mqtt.MQTT_ERR_NO_CONN:
                return 2, mid
            else:
                self.__logger.debug("subscribe error topic:%s" % topic)
                return -1, mid
        # topic format [(topic1, qos),(topic2,qos)]
        if isinstance(topic, list):
            topic_list = []
            for t, qos in topic:
                if qos < 0 or qos > 1:
                    raise ValueError('Invalid QoS level.')
                if t is None or len(t) == 0 or not isinstance(t, str):
                    raise ValueError('Invalid topic.')
                topic_list.append((t, qos))

            rc, mid = self.__mqtt_client.subscribe(topic_list)
            if rc == mqtt.MQTT_ERR_SUCCESS:
                self.__logger.debug("subscribe success topic:%s" % topic)
                return 0, mid
            elif rc == mqtt.MQTT_ERR_NO_CONN:
                return 2, mid
            else:
                self.__logger.debug("subscribe error topic:%s" % topic)
                return -1, mid

    def unsubscribe(self, topic):
        rc, mid = -1, -1
        if self.__mqtt_client is None:
            return rc, mid

        rc, mid = self.__mqtt_client.unsubscribe(topic)
        if rc == mqtt.MQTT_ERR_SUCCESS:
            return 0, mid
        else:
            return rc, mid

    def publish(self, topic, payload, qos):
        rc, mid = -1, -1
        if self.__mqtt_client is None:
            return rc, mid

        # rc, mid = self.__mqtt_client.publish(topic, json.dumps(payload), qos)
        rc, mid = self.__mqtt_client.publish(topic, payload, qos)
        if rc == mqtt.MQTT_ERR_SUCCESS:
            self.__logger.debug("publish success")
            return 0, mid
        else:
            self.__logger.debug("publish failed")
            return rc, mid



#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import hmac
import ssl
import base64
from Crypto.Cipher import AES

class Codec(object):
    def __init__(self):
        self.__init = None

    class _AESUtil:
        __BLOCK_SIZE_16 = BLOCK_SIZE_16 = AES.block_size

        '''
        @staticmethod
        def encryt(str, key, iv):
            cipher = AES.new(key, AES.MODE_CBC, iv)
            x = AESUtil.__BLOCK_SIZE_16 - (len(str) % AESUtil.__BLOCK_SIZE_16)
            if x != 0:
                str = str + chr(x) * x
            msg = cipher.encrypt(str)
            msg = base64.b64encode(msg)
            return msg
        '''

        @staticmethod
        def decrypt(encrypt_str, key, init_vector):
            cipher = AES.new(key, AES.MODE_CBC, init_vector)
            decrypt_bytes = base64.b64decode(encrypt_str)
            return cipher.decrypt(decrypt_bytes)

    class Hmac:
        @staticmethod
        def sha1_encode(key, content):
            return hmac.new(key, content, hashlib.sha1).digest()

        @staticmethod
        def sha256_encode(key, content):
            return hmac.new(key, content, hashlib.sha256).digest()

    class Hash:
        @staticmethod
        def sha256_encode(content):
            return hashlib.sha256(content).hexdigest()

    class Base64:
        @staticmethod
        def encode(content):
            return base64.b64encode(content).decode('utf-8')

    class Ssl():
        __IOT_CA_CRT = "\
-----BEGIN CERTIFICATE-----\n\
MIIDxTCCAq2gAwIBAgIJALM1winYO2xzMA0GCSqGSIb3DQEBCwUAMHkxCzAJBgNV\n\
BAYTAkNOMRIwEAYDVQQIDAlHdWFuZ0RvbmcxETAPBgNVBAcMCFNoZW5aaGVuMRAw\n\
DgYDVQQKDAdUZW5jZW50MRcwFQYDVQQLDA5UZW5jZW50IElvdGh1YjEYMBYGA1UE\n\
AwwPd3d3LnRlbmNlbnQuY29tMB4XDTE3MTEyNzA0MjA1OVoXDTMyMTEyMzA0MjA1\n\
OVoweTELMAkGA1UEBhMCQ04xEjAQBgNVBAgMCUd1YW5nRG9uZzERMA8GA1UEBwwI\n\
U2hlblpoZW4xEDAOBgNVBAoMB1RlbmNlbnQxFzAVBgNVBAsMDlRlbmNlbnQgSW90\n\
aHViMRgwFgYDVQQDDA93d3cudGVuY2VudC5jb20wggEiMA0GCSqGSIb3DQEBAQUA\n\
A4IBDwAwggEKAoIBAQDVxwDZRVkU5WexneBEkdaKs4ehgQbzpbufrWo5Lb5gJ3i0\n\
eukbOB81yAaavb23oiNta4gmMTq2F6/hAFsRv4J2bdTs5SxwEYbiYU1teGHuUQHO\n\
iQsZCdNTJgcikga9JYKWcBjFEnAxKycNsmqsq4AJ0CEyZbo//IYX3czEQtYWHjp7\n\
FJOlPPd1idKtFMVNG6LGXEwS/TPElE+grYOxwB7Anx3iC5ZpE5lo5tTioFTHzqbT\n\
qTN7rbFZRytAPk/JXMTLgO55fldm4JZTP3GQsPzwIh4wNNKhi4yWG1o2u3hAnZDv\n\
UVFV7al2zFdOfuu0KMzuLzrWrK16SPadRDd9eT17AgMBAAGjUDBOMB0GA1UdDgQW\n\
BBQrr48jv4FxdKs3r0BkmJO7zH4ALzAfBgNVHSMEGDAWgBQrr48jv4FxdKs3r0Bk\n\
mJO7zH4ALzAMBgNVHRMEBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBAQDRSjXnBc3T\n\
d9VmtTCuALXrQELY8KtM+cXYYNgtodHsxmrRMpJofsPGiqPfb82klvswpXxPK8Xx\n\
SuUUo74Fo+AEyJxMrRKlbJvlEtnpSilKmG6rO9+bFq3nbeOAfat4lPl0DIscWUx3\n\
ajXtvMCcSwTlF8rPgXbOaSXZidRYNqSyUjC2Q4m93Cv+KlyB+FgOke8x4aKAkf5p\n\
XR8i1BN1OiMTIRYhGSfeZbVRq5kTdvtahiWFZu9DGO+hxDZObYGIxGHWPftrhBKz\n\
RT16Amn780rQLWojr70q7o7QP5tO0wDPfCdFSc6CQFq/ngOzYag0kJ2F+O5U6+kS\n\
QVrcRBDxzx/G\n\
-----END CERTIFICATE-----"

        def __init__(self):
            self.__iot_ca_crt = self.__IOT_CA_CRT

        def create_content(self, cadata=None):
            if cadata is None:
                cadata = self.__iot_ca_crt
            return ssl.create_default_context(ssl.Purpose.CLIENT_AUTH, cadata=cadata)

            

    


#
# Tencent is pleased to support the open source community by making IoT Hub available.
# Copyright (C) 2016 THL A29 Limited, a Tencent company. All rights reserved.

# Licensed under the MIT License (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://opensource.org/licenses/MIT

# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.

import json
from hub.utils.providers import ConnClientProvider
from hub.utils.providers import TopicProvider

class Broadcast(object):
    def __init__(self, host, product_id, device_name, device_secret,
                    websocket=False, tls=True, logger=None):
        self.__provider = ConnClientProvider(host, product_id, device_name, device_secret,
                                                websocket=websocket, tls=tls, logger=logger)
        self.__protocol = self.__provider.protocol
        self.__topic = TopicProvider(product_id, device_name)
        self.__logger = logger
        self.__user_callback = {}

    def __assert(self, param):
        if param is None or len(param) == 0:
            raise ValueError('Invalid param.')

    def handle_broadcast(self, topic, qos, payload, userdata):
        if self.__user_callback[topic] is not None:
                self.__user_callback[topic](topic, qos, payload, userdata)
        else:
            self.__logger.error("no callback for topic %s" % topic)

    def broadcast_init(self, broadcast_cb):
        self.__user_callback[self.__topic.broadcast_topic_sub] = broadcast_cb
        return self.__protocol.subscribe(self.__topic.broadcast_topic_sub, 0)


import json
import re
import urllib.request
import setuptools

name = "tencent-iot-device"
# resp = urllib.request.urlopen(f'https://pypi.org/pypi/{name}/json')
# resp = urllib.request.urlopen(f'https://pypi.org/pypi/{name}/json')
# data = json.loads(resp.read().decode("utf-8"))
# version = data['info']['version']
# last_num = re.findall(r'\d+', version)[-1]
# version = version[:-len(last_num)] + str(int(last_num) + 1)

with open("README.md", mode='r', encoding='UTF-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name=("%s" % name),
    version="1.0.0",
    # version="1.0.0",
    author="larrytin",
    author_email="dev_tester@163.com",
    description="Tencent IoT Explorer SDK for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tencentyun/iot-device-python",
    packages=setuptools.find_packages(),
    install_requires = ["paho-mqtt==1.5.1", "pycrypto==2.6.1"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
